import unittest

from aioprometheus import (
    CollectorRegistry,
    Collector,
    Counter,
    Gauge,
    Summary)


class TestRegistry(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

    def test_register(self):
        ''' check collectors can be registered '''

        q = 100
        collectors = [
            Collector('test' + str(i), 'Test' + str(i)) for i in range(q)]

        r = CollectorRegistry()

        for i in collectors:
            r.register(i)

        self.assertEqual(q, len(r.collectors))

    def test_register_sames(self):
        ''' check registering same metrics raises exceptoion '''
        r = CollectorRegistry()

        r.register(Collector(**self.data))

        with self.assertRaises(ValueError) as context:
            r.register(Collector(**self.data))

        self.assertEqual(
            "Collector {} is already registered".format(self.data['name']),
            str(context.exception))

    def test_register_counter(self):
        ''' check registering a counter collector '''
        r = CollectorRegistry()
        r.register(Counter(**self.data))

        self.assertEqual(1, len(r.collectors))

    def test_register_gauge(self):
        ''' check registering a gauge collector '''
        r = CollectorRegistry()
        r.register(Gauge(**self.data))

        self.assertEqual(1, len(r.collectors))

    def test_register_summary(self):
        ''' check registering a summary collector '''
        r = CollectorRegistry()
        r.register(Summary(**self.data))

        self.assertEqual(1, len(r.collectors))

    def test_register_wrong_type(self):
        ''' check registering an invalid collector raises an exception '''
        r = CollectorRegistry()

        with self.assertRaises(TypeError) as context:
            r.register("This will fail")
        self.assertIn(
            "Invalid collector type: ",
            str(context.exception))

    def test_deregister(self):
        ''' check collectors can be deregistered '''
        r = CollectorRegistry()
        r.register(Collector(**self.data))

        r.deregister(self.data['name'])

        self.assertEqual(0, len(r.collectors))

    def test_get(self):
        r = CollectorRegistry()
        c = Collector(**self.data)
        r.register(c)

        self.assertEqual(c, r.get(c.name))

    def test_get_all(self):
        q = 100
        collectors = [
            Collector('test' + str(i), 'Test' + str(i)) for i in range(q)]

        r = CollectorRegistry()

        for i in collectors:
            r.register(i)

        result = r.get_all()

        self.assertTrue(isinstance(result, list))
        self.assertEqual(q, len(result))



'''
This module implements an asynchronous Prometheus.io metrics service.
The service is built upon ``aiohttp``.
'''

import asyncio
import logging

import aiohttp
import aiohttp.web

from aiohttp.hdrs import (
    METH_GET as GET,
    ACCEPT)

from .negotiator import negotiate
from .registry import Registry
from typing import Set

# imports only used for type annotations
from asyncio.base_events import BaseEventLoop, Server
from ssl import SSLContext


logger = logging.getLogger(__name__)

DEFAULT_METRICS_PATH = '/metrics'


class Service(object):
    '''
    This class implements a Prometheus.io metrics service that can
    be embedded within asyncio based applications so they can be scraped
    by the Prometheus.io server.
    '''

    def __init__(self,
                 registry: Registry = None,
                 loop: BaseEventLoop = None) -> None:
        '''
        Initialise the Prometheus metrics service.

        :param registry: A :class:`CollectorRegistry` instance that will
          hold all the metrics for this service. If no registry if specified
          then the default registry is used.

        :param loop: The event loop instance to use. If no loop is specified
          then the default event loop will be retrieved.

        :raises: Exception if the registry object is not an instance of the
          Registry type.
        '''
        self.loop = loop or asyncio.get_event_loop()
        if registry is not None and not isinstance(registry, Registry):
            raise Exception(
                'registry must be a Registry, got: {}'.format(registry))
        self.registry = registry or Registry()
        self._svr = None  # type: Server
        self._svc = None  # type: aiohttp.web.Application
        self._handler = None  # type: aiohttp.web.RequestHandlerFactory
        self._https = None  # type: bool

    @property
    def url(self) -> str:
        ''' Return the Prometheus metrics url

        :raises: Exception if the server has not been started.

        :return: the service URL as a string
        '''
        if self._svr is None:
            raise Exception(
                "No URL available, Prometheus metrics server is not running")

        # IPv4 returns 2-tuple, IPv6 returns 4-tuple
        host, port, *_ = self._svr.sockets[0].getsockname()
        scheme = "http{}".format('s' if self._https else '')
        url = "{scheme}://{host}:{port}{metrics_url}".format(
            scheme=scheme,
            host=host if ":" not in host else "[{}]".format(host),
            port=port,
            metrics_url=self._metrics_url)
        return url

    async def start(self,
                    addr: str = '',
                    port: int = 0,
                    ssl: SSLContext = None,
                    metrics_url: str = DEFAULT_METRICS_PATH,
                    discovery_agent=None) -> None:
        ''' Start the prometheus metrics HTTP(S) server.

        :param addr: the address to bind the server on. By default this is
          set to an empty string so that the service becomes available on
          all interfaces.

        :param port: The port to bind the server on. The default value is 0
          which will cause the server to bind to an ephemeral port. If you
          want the server to operate on a fixed port then you need to specifiy
          the port.

        :param ssl: a sslContext for use with TLS.

        :param metrics_url: The name of the endpoint route to expose
          prometheus metrics on. Defaults to '/metrics'.

        :param discovery_agent: an agent that can register the metrics
          service with a service discovery mechanism.

        :raises: Exception if the server could not be started.
        '''
        logger.debug(
            'Prometheus metrics server starting on %s:%s%s',
            addr, port, metrics_url)

        if self._svr:
            logger.warning(
                'Prometheus metrics server is already running')
            return

        self._svc = aiohttp.web.Application()
        self._metrics_url = metrics_url
        self._svc.router.add_route(
            GET, metrics_url, self.handle_metrics)
        self._handler = self._svc.make_handler()
        self._https = ssl is not None
        try:
            self._svr = await self.loop.create_server(
                self._handler, addr, port, ssl=ssl)
        except Exception:
            logger.exception('error creating metrics server')
            raise

        logger.debug('Prometheus metrics server started on %s', self.url)

        # register service with service discovery
        if discovery_agent:
            await discovery_agent.register(self)

    async def stop(self,
                   wait_duration: float = 1.0,
                   discovery_agent=None) -> None:
        ''' Stop the prometheus metrics HTTP(S) server.

        :param wait_duration: the number of seconds to wait for connections to
          finish.

        :param discovery_agent: an agent that can register the metrics
          service with a service discovery mechanism.

        '''
        logger.debug(
            'Prometheus metrics server stopping')

        if self._svr is None:
            logger.warning(
                'Metrics HTTP Server is already stopped')
            return

        # de-register service with service discovery
        if discovery_agent:
            await discovery_agent.deregister(self)

        self._svr.close()
        await self._svr.wait_closed()
        await self._svc.shutdown()
        await self._handler.finish_connections(wait_duration)
        await self._svc.cleanup()
        self._svr = None
        self._svc = None
        self._handler = None
        logger.debug('Prometheus metrics server stopped')

    async def handle_metrics(self,
                             request: aiohttp.web.Request) -> aiohttp.web.Response:
        ''' Handle a request to the metrics route.

        The request is inspected and the most efficient response data format
        is chosen.
        '''
        Formatter = negotiate(self.accepts(request))
        formatter = Formatter(False)

        resp = aiohttp.web.Response()
        resp.headers.update(formatter.get_headers())
        resp.body = formatter.marshall(self.registry)
        return resp

    def accepts(self,
                request: aiohttp.web.Request) -> Set[str]:
        ''' Return a sequence of accepts items in the request headers '''
        accepts = set()  # type: Set[str]
        accept_headers = request.headers.getall(ACCEPT)
        logger.debug('accept: {}'.format(accept_headers))
        for accept_items in accept_headers:
            if ';' in accept_items:
                accept_items = [i.strip() for i in accept_items.split(';')]
            else:
                accept_items = [accept_items]
            accepts.update(accept_items)
        return accepts



import collections
import json
import re


# Sometimes python will access by string for example iterating objects, and
# it has this notation
regex = re.compile("\{.*:.*,?\}")


# http://stackoverflow.com/questions/3387691/python-how-to-perfectly-override-a-dict
class MetricDict(collections.MutableMapping):
    '''
    MetricDict stores the data based on the labels so we need to generate
    custom hash keys based on the labels
    '''

    EMPTY_KEY = "__EMPTY__"

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):

        # Sometimes we need empty keys
        if not key or key == MetricDict.EMPTY_KEY:
            return MetricDict.EMPTY_KEY

        # Python accesses by string key so we allow if is str and
        # 'our custom' format
        if type(key) == str and regex.match(key):
            return key

        if type(key) is not dict:
            raise TypeError("Only accepts dicts as keys")

        return json.dumps(key, sort_keys=True)




from .collectors import (
    Collector, Counter, Gauge, Summary, Histogram)
from .decorators import count_exceptions, inprogress, timer
from .formats import (BinaryFormatter, TextFormatter)
from .pusher import Pusher
from .registry import Registry, CollectorRegistry
from .service import Service

__version__ = "16.08.07"




from aioprometheus import (
    Counter,
    Gauge,
    Histogram,
    Summary,
    timer,
    inprogress,
    count_exceptions)
from aioprometheus.test_utils import AsyncioTestCase


class TestDecorators(AsyncioTestCase):

    async def test_timer(self):
        m = Summary('metric_label', 'metric help')

        # decorator should work methods as well as functions
        @timer(m, {'kind': 'function'})
        async def a():
            return

        await a()

        m_function = m.get({'kind': 'function'})
        self.assertEqual(m_function['count'], 1)

        # decorator should work methods as well as functions
        class B(object):

            @timer(m, {'kind': 'method'})
            async def b(self, arg1, arg2=None):
                return arg1 == 'b_arg', arg2 == 'arg_2'

        b = B()
        results = await b.b('b_arg', arg2='arg_2')
        self.assertTrue(all(results))

        m_method = m.get({'kind': 'method'})
        self.assertEqual(m_method['count'], 1)

        # Only Summary metric type can be used with @timer, others should
        # raise an exception.
        with self.assertRaises(Exception) as cm:
            m = Counter('metric_label', 'metric help')

            @timer(m)
            async def c():
                return
        self.assertIn(
            "timer decorator expects a Summary metric but got:",
            str(cm.exception))

    async def test_inprogress(self):
        m = Gauge('metric_label', 'metric help')

        # decorator should work methods as well as functions
        @inprogress(m, {'kind': 'function'})
        async def a():
            return

        await a()

        m_function_value = m.get({'kind': 'function'})
        self.assertEqual(m_function_value, 0)

        # decorator should work methods as well as functions
        class B(object):

            @inprogress(m, {'kind': 'method'})
            async def b(self, arg1, arg2=None):
                return arg1 == 'b_arg', arg2 == 'arg_2'

        b = B()
        results = await b.b('b_arg', arg2='arg_2')
        self.assertTrue(all(results))

        m_method_value = m.get({'kind': 'method'})
        self.assertEqual(m_method_value, 0)

        # Only Gauge metric type can be used with @timer, others should
        # raise an exception.
        with self.assertRaises(Exception) as cm:
            m = Counter('metric_label', 'metric help')

            @inprogress(m)
            async def c():
                return
        self.assertIn(
            "inprogess decorator expects a Gauge metric but got:",
            str(cm.exception))

    async def test_count_exceptions(self):
        m = Counter('metric_label', 'metric help')

        # decorator should work methods as well as functions
        @count_exceptions(m, {'kind': 'function'})
        async def a(raise_exc=False):
            if raise_exc:
                raise Exception('dummy exception')
            return

        await a()

        # metric should not exist until an exception occurs
        with self.assertRaises(KeyError):
            m.get({'kind': 'function'})

        with self.assertRaises(Exception) as cm:
            await a(True)
        self.assertIn("dummy exception", str(cm.exception))

        m_function_value = m.get({'kind': 'function'})
        self.assertEqual(m_function_value, 1)

        # decorator should work methods as well as functions
        class B(object):

            @count_exceptions(m, {'kind': 'method'})
            async def b(self, arg1, arg2=None, raise_exc=False):
                if raise_exc:
                    raise Exception('dummy exception')
                return arg1 == 'b_arg', arg2 == 'arg_2'

        b = B()

        results = await b.b('b_arg', arg2='arg_2')
        self.assertTrue(all(results))
        # metric should not exist until an exception occurs
        with self.assertRaises(KeyError):
            m.get({'kind': 'method'})

        with self.assertRaises(Exception) as cm:
            await b.b('b_arg', arg2='arg_2', raise_exc=True)
        self.assertIn("dummy exception", str(cm.exception))
        m_method_value = m.get({'kind': 'method'})
        self.assertEqual(m_method_value, 1)

        # Only Gauge metric type can be used with @timer, others should
        # raise an exception.
        with self.assertRaises(Exception) as cm:
            m = Histogram('metric_label', 'metric help')

            @count_exceptions(m)
            async def c():
                return
        self.assertIn(
            "count_exceptions decorator expects a Counter metric but got:",
            str(cm.exception))



''' This module implements a Prometheus metrics binary formatter.

Protocol Buffers is used as the data serialisation format.
'''

import pyrobuf_util
import prometheus_metrics_proto as pmp  # type: ignore

from .base import IFormatter
from ..collectors import Counter, Gauge, Summary, Histogram
from typing import cast, Callable, Dict, List, Tuple, Union

# imports only used for type annotations
from ..registry import CollectorRegistry

# typing aliases
LabelsType = Dict[str, str]
NumericValueType = Union[int, float]
SummaryDictKeyType = Union[float, str]  # e.g. sum, 0.25, etc
SummaryDictType = Dict[SummaryDictKeyType, NumericValueType]
HistogramDictKeyType = Union[float, str]  # e.g. sum, 0.25, etc
HistogramDictType = Dict[HistogramDictKeyType, NumericValueType]
CollectorsType = Union[Counter, Gauge, Histogram, Summary]
MetricValueType = Union[float, SummaryDictType, HistogramDictType]
MetricTupleType = Tuple[LabelsType, MetricValueType]
FormatterFuncType = Callable[[MetricTupleType, str, LabelsType], pmp.Metric]


class BinaryFormatter(IFormatter):
    ''' This formatter encodes into the Protocol Buffers binary format '''

    def __init__(self, timestamp: bool = False) -> None:
        '''
        :param timestamp: a boolean flag that when True will add a timestamp
          to metric.
        '''
        self.timestamp = timestamp
        self._headers = {
            'Content-Type': "application/vnd.google.protobuf; "
                            "proto=io.prometheus.client.MetricFamily; "
                            "encoding=delimited"}

    def get_headers(self) -> Dict[str, str]:
        return self._headers

    def _create_labels(self, labels: LabelsType) -> List[pmp.LabelPair]:
        ''' Return a list of LabelPair objects for each label. '''
        return [
            pmp.LabelPair(name=k, value=str(v)) for k, v in labels.items()]

    def _format_counter(self,
                        # counter: Tuple[LabelsType, ValueType],
                        counter: MetricTupleType,
                        name: str,
                        const_labels: LabelsType) -> pmp.Metric:
        '''
        :param counter: a 2-tuple containing labels and the counter value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        counter_labels, counter_value = counter
        labels = self._unify_labels(counter_labels, const_labels, ordered=True)
        pb_labels = self._create_labels(labels)

        pb_counter = pmp.Counter(value=counter_value)

        pb_metric = pmp.Metric(label=pb_labels, counter=pb_counter)
        if self.timestamp:
            pb_metric.timestamp_ms = self._get_timestamp()

        return pb_metric

    def _format_gauge(self,
                      gauge: MetricTupleType,
                      name: str,
                      const_labels: LabelsType) -> pmp.Metric:
        '''
        :param gauge: a 2-tuple containing labels and the gauge value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        gauge_labels, gauge_value = gauge
        labels = self._unify_labels(gauge_labels, const_labels, ordered=True)
        pb_labels = self._create_labels(labels)

        pb_gauge = pmp.Gauge(value=gauge_value)

        pb_metric = pmp.Metric(label=pb_labels, gauge=pb_gauge)
        if self.timestamp:
            pb_metric.timestamp_ms = self._get_timestamp()
        return pb_metric

    def _format_summary(self,
                        summary: MetricTupleType,
                        name: str,
                        const_labels: LabelsType) -> pmp.Metric:
        '''
        :param summary: a 2-tuple containing labels and a dict representing
          the summary value. The dict contains keys for each quantile as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''

        summary_labels, summary_value_dict = summary
        # typing check, no runtime behaviour.
        summary_value_dict = cast(SummaryDictType, summary_value_dict)
        labels = self._unify_labels(summary_labels, const_labels, ordered=True)
        pb_labels = self._create_labels(labels)

        pb_quantiles = []
        for k, v in summary_value_dict.items():
            if not isinstance(k, str):
                pb_quantile = pmp.Quantile(quantile=k, value=v)
                pb_quantiles.append(pb_quantile)

        pb_summary = pmp.Summary(
            sample_count=summary_value_dict['count'],
            sample_sum=summary_value_dict['sum'],
            quantile=pb_quantiles)

        pb_metric = pmp.Metric(label=pb_labels, summary=pb_summary)
        if self.timestamp:
            pb_metric.timestamp_ms = self._get_timestamp()

        return pb_metric

    def _format_histogram(self,
                          histogram: MetricTupleType,
                          name: str,
                          const_labels: LabelsType) -> pmp.Metric:
        '''
        :param histogram: a 2-tuple containing labels and a dict representing
          the histogram value. The dict contains keys for each bucket as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        histogram_labels, histogram_value_dict = histogram
        # typing check, no runtime behaviour.
        histogram_value_dict = cast(HistogramDictType, histogram_value_dict)
        labels = self._unify_labels(
            histogram_labels, const_labels, ordered=True)
        pb_labels = self._create_labels(labels)

        pb_buckets = []
        for k, v in histogram_value_dict.items():
            if not isinstance(k, str):
                pb_bucket = pmp.Bucket(cumulative_count=v, upper_bound=k)
                pb_buckets.append(pb_bucket)

        pb_histogram = pmp.Histogram(
            sample_count=histogram_value_dict['count'],
            sample_sum=histogram_value_dict['sum'],
            bucket=pb_buckets)

        pb_metric = pmp.Metric(label=pb_labels, histogram=pb_histogram)
        if self.timestamp:
            pb_metric.timestamp_ms = self._get_timestamp()

        return pb_metric

    def marshall_collector(self,
                           collector: CollectorsType) -> pmp.MetricFamily:
        '''
        Marshalls a collector into :class:`MetricFamily` object representing
        the metrics in the collector.

        :return: a :class:`MetricFamily` object
        '''
        exec_method = None  # type: FormatterFuncType
        if isinstance(collector, Counter):
            metric_type = pmp.COUNTER
            exec_method = self._format_counter
        elif isinstance(collector, Gauge):
            metric_type = pmp.GAUGE
            exec_method = self._format_gauge
        elif isinstance(collector, Summary):
            metric_type = pmp.SUMMARY
            exec_method = self._format_summary
        elif isinstance(collector, Histogram):
            metric_type = pmp.HISTOGRAM
            exec_method = self._format_histogram
        else:
            raise TypeError("Not a valid object format")

        pb_metrics = []

        for i in collector.get_all():
            i = cast(MetricTupleType, i)  # typing check, no runtime behaviour.
            r = exec_method(i, collector.name, collector.const_labels)
            pb_metrics.append(r)

        pb_metric_family = pmp.MetricFamily(
            name=collector.name, help=collector.doc,
            type=metric_type, metric=pb_metrics)

        return pb_metric_family

    def marshall(self, registry: CollectorRegistry) -> bytes:
        ''' Marshall the collectors in the registry into binary protocol
        buffer format.

        The Prometheus metrics parser expects each metric (MetricFamily) to
        be prefixed with a varint containing the size of the encoded metric.

        :returns: bytes
        '''
        payload = []
        for i in registry.get_all():
            encoded_metric = self.marshall_collector(i).SerializeToString()
            length = pyrobuf_util.to_varint(len(encoded_metric))
            payload.append(length + encoded_metric)
        return b"".join(payload)



import datetime
import re
import time
import unittest

from aioprometheus import (
    Collector, Counter, Gauge, Histogram, Summary,
    TextFormatter, BinaryFormatter,
    Registry)
import prometheus_metrics_proto as pmp


class TestTextFormat(unittest.TestCase):

    def test_headers(self):
        f = TextFormatter()
        result = {
            'Content-Type': 'text/plain; version=0.0.4; charset=utf-8'
        }

        self.assertEqual(result, f.get_headers())

    def test_wrong_format(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

        f = TextFormatter()

        c = Collector(**self.data)

        with self.assertRaises(TypeError) as context:
            f.marshall_collector(c)

        self.assertEqual('Not a valid object format', str(context.exception))

    def test_counter_format(self):

        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }
        c = Counter(**self.data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
            ({'country': "ch", "device": "mobile"}, 654),
            ({'country': "ca", "device": "desktop"}, 1001),
            ({'country': "jp", "device": "desktop"}, 995),
            ({'country': "au", "device": "desktop"}, 520),
            ({'country': "py", "device": "mobile"}, 654),
            ({'country': "ar", "device": "desktop"}, 1001),
            ({'country': "pt", "device": "desktop"}, 995),
        )

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total counter",
            "logged_users_total{country=\"ch\",device=\"mobile\"} 654",
            "logged_users_total{country=\"zh\",device=\"desktop\"} 520",
            "logged_users_total{country=\"jp\",device=\"desktop\"} 995",
            "logged_users_total{country=\"de\",device=\"desktop\"} 995",
            "logged_users_total{country=\"pt\",device=\"desktop\"} 995",
            "logged_users_total{country=\"ca\",device=\"desktop\"} 1001",
            "logged_users_total{country=\"sp\",device=\"desktop\"} 520",
            "logged_users_total{country=\"au\",device=\"desktop\"} 520",
            "logged_users_total{country=\"uk\",device=\"desktop\"} 1001",
            "logged_users_total{country=\"py\",device=\"mobile\"} 654",
            "logged_users_total{country=\"us\",device=\"mobile\"} 654",
            "logged_users_total{country=\"ar\",device=\"desktop\"} 1001",
        )

        # Add data to the collector
        for i in counter_data:
            c.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()
        result = f.marshall_lines(c)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_counter_format_with_const_labels(self):

        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }
        c = Counter(**self.data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
            ({'country': "ch", "device": "mobile"}, 654),
            ({'country': "ca", "device": "desktop"}, 1001),
            ({'country': "jp", "device": "desktop"}, 995),
            ({'country': "au", "device": "desktop"}, 520),
            ({'country': "py", "device": "mobile"}, 654),
            ({'country': "ar", "device": "desktop"}, 1001),
            ({'country': "pt", "device": "desktop"}, 995),
        )

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total counter",
            "logged_users_total{app=\"my_app\",country=\"ch\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"zh\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"jp\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"de\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"pt\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"ca\",device=\"desktop\"} 1001",
            "logged_users_total{app=\"my_app\",country=\"sp\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"au\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"uk\",device=\"desktop\"} 1001",
            "logged_users_total{app=\"my_app\",country=\"py\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"us\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"ar\",device=\"desktop\"} 1001",
        )

        # Add data to the collector
        for i in counter_data:
            c.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()
        result = f.marshall_lines(c)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_counter_format_text(self):

        name = "container_cpu_usage_seconds_total"
        doc = "Total seconds of cpu time consumed."

        valid_result = """# HELP container_cpu_usage_seconds_total Total seconds of cpu time consumed.
# TYPE container_cpu_usage_seconds_total counter
container_cpu_usage_seconds_total{id="110863b5395f7f3476d44e7cb8799f2643abbd385dd544bcc379538ac6ffc5ca",name="container-extractor",type="kernel"} 0
container_cpu_usage_seconds_total{id="110863b5395f7f3476d44e7cb8799f2643abbd385dd544bcc379538ac6ffc5ca",name="container-extractor",type="user"} 0
container_cpu_usage_seconds_total{id="7c1ae8f404be413a6413d0792123092446f694887f52ae6403356215943d3c36",name="calendall_db_1",type="kernel"} 0
container_cpu_usage_seconds_total{id="7c1ae8f404be413a6413d0792123092446f694887f52ae6403356215943d3c36",name="calendall_db_1",type="user"} 0
container_cpu_usage_seconds_total{id="c863b092d1ecdc68f54a6a4ed0d24fe629696be2337ccafb44c279c7c2d1c172",name="calendall_web_run_8",type="kernel"} 0
container_cpu_usage_seconds_total{id="c863b092d1ecdc68f54a6a4ed0d24fe629696be2337ccafb44c279c7c2d1c172",name="calendall_web_run_8",type="user"} 0
container_cpu_usage_seconds_total{id="cefa0b389a634a0b2f3c2f52ade668d71de75e5775e91297bd65bebe745ba054",name="prometheus",type="kernel"} 0
container_cpu_usage_seconds_total{id="cefa0b389a634a0b2f3c2f52ade668d71de75e5775e91297bd65bebe745ba054",name="prometheus",type="user"} 0"""

        data = (
            ({'id': "110863b5395f7f3476d44e7cb8799f2643abbd385dd544bcc379538ac6ffc5ca", 'name': "container-extractor", 'type': "kernel"}, 0),
            ({'id': "110863b5395f7f3476d44e7cb8799f2643abbd385dd544bcc379538ac6ffc5ca", 'name': "container-extractor", 'type': "user"}, 0),
            ({'id': "7c1ae8f404be413a6413d0792123092446f694887f52ae6403356215943d3c36", 'name': "calendall_db_1", 'type': "kernel"}, 0),
            ({'id': "7c1ae8f404be413a6413d0792123092446f694887f52ae6403356215943d3c36", 'name': "calendall_db_1", 'type': "user"}, 0),
            ({'id': "c863b092d1ecdc68f54a6a4ed0d24fe629696be2337ccafb44c279c7c2d1c172", 'name': "calendall_web_run_8", 'type': "kernel"}, 0),
            ({'id': "c863b092d1ecdc68f54a6a4ed0d24fe629696be2337ccafb44c279c7c2d1c172", 'name': "calendall_web_run_8", 'type': "user"}, 0),
            ({'id': "cefa0b389a634a0b2f3c2f52ade668d71de75e5775e91297bd65bebe745ba054", 'name': "prometheus", 'type': "kernel"}, 0),
            ({'id': "cefa0b389a634a0b2f3c2f52ade668d71de75e5775e91297bd65bebe745ba054", 'name': "prometheus", 'type': "user"}, 0),
        )

        # Create the counter
        c = Counter(name=name, doc=doc, const_labels={})

        for i in data:
            c.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()

        result = f.marshall_collector(c)

        self.assertEqual(valid_result, result)

    def test_counter_format_with_timestamp(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }
        c = Counter(**self.data)

        counter_data = ({'country': "ch", "device": "mobile"}, 654)

        c.set_value(counter_data[0], counter_data[1])

        result_regex = """# HELP logged_users_total Logged users in the application
# TYPE logged_users_total counter
logged_users_total{country="ch",device="mobile"} 654 \d*(?:.\d*)?$"""

        f_with_ts = TextFormatter(True)
        result = f_with_ts.marshall_collector(c)

        self.assertTrue(re.match(result_regex, result))

    def test_single_counter_format_text(self):

        name = "prometheus_dns_sd_lookups_total"
        doc = "The number of DNS-SD lookups."

        valid_result = """# HELP prometheus_dns_sd_lookups_total The number of DNS-SD lookups.
# TYPE prometheus_dns_sd_lookups_total counter
prometheus_dns_sd_lookups_total 10"""

        data = (
            (None, 10),
        )

        # Create the counter
        c = Counter(name=name, doc=doc, const_labels={})

        for i in data:
            c.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()

        result = f.marshall_collector(c)

        self.assertEqual(valid_result, result)

    def test_gauge_format(self):

        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }
        g = Gauge(**self.data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
            ({'country': "ch", "device": "mobile"}, 654),
            ({'country': "ca", "device": "desktop"}, 1001),
            ({'country': "jp", "device": "desktop"}, 995),
            ({'country': "au", "device": "desktop"}, 520),
            ({'country': "py", "device": "mobile"}, 654),
            ({'country': "ar", "device": "desktop"}, 1001),
            ({'country': "pt", "device": "desktop"}, 995),
        )

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total gauge",
            "logged_users_total{country=\"ch\",device=\"mobile\"} 654",
            "logged_users_total{country=\"zh\",device=\"desktop\"} 520",
            "logged_users_total{country=\"jp\",device=\"desktop\"} 995",
            "logged_users_total{country=\"de\",device=\"desktop\"} 995",
            "logged_users_total{country=\"pt\",device=\"desktop\"} 995",
            "logged_users_total{country=\"ca\",device=\"desktop\"} 1001",
            "logged_users_total{country=\"sp\",device=\"desktop\"} 520",
            "logged_users_total{country=\"au\",device=\"desktop\"} 520",
            "logged_users_total{country=\"uk\",device=\"desktop\"} 1001",
            "logged_users_total{country=\"py\",device=\"mobile\"} 654",
            "logged_users_total{country=\"us\",device=\"mobile\"} 654",
            "logged_users_total{country=\"ar\",device=\"desktop\"} 1001",
        )

        # Add data to the collector
        for i in counter_data:
            g.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()
        result = f.marshall_lines(g)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_gauge_format_with_const_labels(self):

        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }
        g = Gauge(**self.data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
            ({'country': "ch", "device": "mobile"}, 654),
            ({'country': "ca", "device": "desktop"}, 1001),
            ({'country': "jp", "device": "desktop"}, 995),
            ({'country': "au", "device": "desktop"}, 520),
            ({'country': "py", "device": "mobile"}, 654),
            ({'country': "ar", "device": "desktop"}, 1001),
            ({'country': "pt", "device": "desktop"}, 995),
        )

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total gauge",
            "logged_users_total{app=\"my_app\",country=\"ch\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"zh\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"jp\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"de\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"pt\",device=\"desktop\"} 995",
            "logged_users_total{app=\"my_app\",country=\"ca\",device=\"desktop\"} 1001",
            "logged_users_total{app=\"my_app\",country=\"sp\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"au\",device=\"desktop\"} 520",
            "logged_users_total{app=\"my_app\",country=\"uk\",device=\"desktop\"} 1001",
            "logged_users_total{app=\"my_app\",country=\"py\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"us\",device=\"mobile\"} 654",
            "logged_users_total{app=\"my_app\",country=\"ar\",device=\"desktop\"} 1001",
        )

        # Add data to the collector
        for i in counter_data:
            g.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()
        result = f.marshall_lines(g)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_gauge_format_text(self):

        name = "container_memory_max_usage_bytes"
        doc = "Maximum memory usage ever recorded in bytes."

        valid_result = """# HELP container_memory_max_usage_bytes Maximum memory usage ever recorded in bytes.
# TYPE container_memory_max_usage_bytes gauge
container_memory_max_usage_bytes{id="4f70875bb57986783064fe958f694c9e225643b0d18e9cde6bdee56d47b7ce76",name="prometheus"} 0
container_memory_max_usage_bytes{id="89042838f24f0ec0aa2a6c93ff44fd3f3e43057d35cfd32de89558112ecb92a0",name="calendall_web_run_3"} 0
container_memory_max_usage_bytes{id="d11c6bc95459822e995fac4d4ae527f6cac442a1896a771dbb307ba276beceb9",name="db"} 0
container_memory_max_usage_bytes{id="e4260cc9dca3e4e50ad2bffb0ec7432442197f135023ab629fe3576485cc65dd",name="container-extractor"} 0
container_memory_max_usage_bytes{id="f30d1caaa142b1688a0684ed744fcae6d202a36877617b985e20a5d33801b311",name="calendall_db_1"} 0
container_memory_max_usage_bytes{id="f835d921ffaf332f8d88ef5231ba149e389a2f37276f081878d6f982ef89a981",name="cocky_fermat"} 0"""

        data = (
            ({'id': "4f70875bb57986783064fe958f694c9e225643b0d18e9cde6bdee56d47b7ce76", 'name': "prometheus"}, 0),
            ({'id': "89042838f24f0ec0aa2a6c93ff44fd3f3e43057d35cfd32de89558112ecb92a0", 'name': "calendall_web_run_3"}, 0),
            ({'id': "d11c6bc95459822e995fac4d4ae527f6cac442a1896a771dbb307ba276beceb9", 'name': "db"}, 0),
            ({'id': "e4260cc9dca3e4e50ad2bffb0ec7432442197f135023ab629fe3576485cc65dd", 'name': "container-extractor"}, 0),
            ({'id': "f30d1caaa142b1688a0684ed744fcae6d202a36877617b985e20a5d33801b311", 'name': "calendall_db_1"}, 0),
            ({'id': "f835d921ffaf332f8d88ef5231ba149e389a2f37276f081878d6f982ef89a981", 'name': "cocky_fermat"}, 0),
        )

        # Create the counter
        g = Gauge(name=name, doc=doc, const_labels={})

        for i in data:
            g.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()

        result = f.marshall_collector(g)

        self.assertEqual(valid_result, result)

    def test_gauge_format_with_timestamp(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }
        g = Gauge(**self.data)

        counter_data = ({'country': "ch", "device": "mobile"}, 654)

        g.set_value(counter_data[0], counter_data[1])

        result_regex = """# HELP logged_users_total Logged users in the application
# TYPE logged_users_total gauge
logged_users_total{country="ch",device="mobile"} 654 \d*(?:.\d*)?$"""

        f_with_ts = TextFormatter(True)
        result = f_with_ts.marshall_collector(g)

        self.assertTrue(re.match(result_regex, result))

    def test_single_gauge_format_text(self):

        name = "prometheus_local_storage_indexing_queue_capacity"
        doc = "The capacity of the indexing queue."

        valid_result = """# HELP prometheus_local_storage_indexing_queue_capacity The capacity of the indexing queue.
# TYPE prometheus_local_storage_indexing_queue_capacity gauge
prometheus_local_storage_indexing_queue_capacity 16384"""

        data = (
            (None, 16384),
        )

        # Create the counter
        g = Gauge(name=name, doc=doc, const_labels={})

        for i in data:
            g.set_value(i[0], i[1])

        # Select format
        f = TextFormatter()

        result = f.marshall_collector(g)

        self.assertEqual(valid_result, result)

    def test_one_summary_format(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }

        labels = {'handler': '/static'}
        values = [3, 5.2, 13, 4]

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total summary",
            "logged_users_total{handler=\"/static\",quantile=\"0.5\"} 4.0",
            "logged_users_total{handler=\"/static\",quantile=\"0.9\"} 5.2",
            "logged_users_total{handler=\"/static\",quantile=\"0.99\"} 5.2",
            "logged_users_total_count{handler=\"/static\"} 4",
            "logged_users_total_sum{handler=\"/static\"} 25.2",
        )

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        f = TextFormatter()
        result = f.marshall_lines(s)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_summary_format_text(self):
        data = {
            'name': "prometheus_target_interval_length_seconds",
            'doc': "Actual intervals between scrapes.",
            'const_labels': {},
        }

        labels = {'interval': '5s'}
        values = [3, 5.2, 13, 4]

        valid_result = """# HELP prometheus_target_interval_length_seconds Actual intervals between scrapes.
# TYPE prometheus_target_interval_length_seconds summary
prometheus_target_interval_length_seconds_count{interval="5s"} 4
prometheus_target_interval_length_seconds_sum{interval="5s"} 25.2
prometheus_target_interval_length_seconds{interval="5s",quantile="0.5"} 4.0
prometheus_target_interval_length_seconds{interval="5s",quantile="0.9"} 5.2
prometheus_target_interval_length_seconds{interval="5s",quantile="0.99"} 5.2"""

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        f = TextFormatter()
        result = f.marshall_collector(s)

        self.assertEqual(valid_result, result)

    def test_multiple_summary_format(self):
        data = {
            'name': "prometheus_target_interval_length_seconds",
            'doc': "Actual intervals between scrapes.",
            'const_labels': {},
        }

        summary_data = (
            ({'interval': "5s"}, [3, 5.2, 13, 4]),
            ({'interval': "10s"}, [1.3, 1.2, 32.1, 59.2, 109.46, 70.9]),
            ({'interval': "10s", 'method': "fast"}, [5, 9.8, 31, 9.7, 101.4]),
        )

        valid_result = (
            "# HELP prometheus_target_interval_length_seconds Actual intervals between scrapes.",
            "# TYPE prometheus_target_interval_length_seconds summary",
            "prometheus_target_interval_length_seconds{interval=\"5s\",quantile=\"0.5\"} 4.0",
            "prometheus_target_interval_length_seconds{interval=\"5s\",quantile=\"0.9\"} 5.2",
            "prometheus_target_interval_length_seconds{interval=\"5s\",quantile=\"0.99\"} 5.2",
            "prometheus_target_interval_length_seconds_count{interval=\"5s\"} 4",
            "prometheus_target_interval_length_seconds_sum{interval=\"5s\"} 25.2",
            "prometheus_target_interval_length_seconds{interval=\"10s\",quantile=\"0.5\"} 32.1",
            "prometheus_target_interval_length_seconds{interval=\"10s\",quantile=\"0.9\"} 59.2",
            "prometheus_target_interval_length_seconds{interval=\"10s\",quantile=\"0.99\"} 59.2",
            "prometheus_target_interval_length_seconds_count{interval=\"10s\"} 6",
            "prometheus_target_interval_length_seconds_sum{interval=\"10s\"} 274.15999999999997",
            "prometheus_target_interval_length_seconds{interval=\"10s\",method=\"fast\",quantile=\"0.5\"} 9.7",
            "prometheus_target_interval_length_seconds{interval=\"10s\",method=\"fast\",quantile=\"0.9\"} 9.8",
            "prometheus_target_interval_length_seconds{interval=\"10s\",method=\"fast\",quantile=\"0.99\"} 9.8",
            "prometheus_target_interval_length_seconds_count{interval=\"10s\",method=\"fast\"} 5",
            "prometheus_target_interval_length_seconds_sum{interval=\"10s\",method=\"fast\"} 156.9",
        )

        s = Summary(**data)

        for i in summary_data:
            for j in i[1]:
                s.add(i[0], j)

        f = TextFormatter()
        result = f.marshall_lines(s)

        self.assertEqual(sorted(valid_result), sorted(result))

#    # This one hans't got labels
    def test_single_summary_format(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }

        labels = {}
        values = [3, 5.2, 13, 4]

        valid_result = (
            "# HELP logged_users_total Logged users in the application",
            "# TYPE logged_users_total summary",
            "logged_users_total{quantile=\"0.5\"} 4.0",
            "logged_users_total{quantile=\"0.9\"} 5.2",
            "logged_users_total{quantile=\"0.99\"} 5.2",
            "logged_users_total_count 4",
            "logged_users_total_sum 25.2",
        )

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        f = TextFormatter()
        result = f.marshall_lines(s)

        result = sorted(result)
        valid_result = sorted(valid_result)

        self.assertEqual(valid_result, result)

    def test_summary_format_timestamp(self):
        data = {
            'name': "prometheus_target_interval_length_seconds",
            'doc': "Actual intervals between scrapes.",
            'const_labels': {},
        }

        labels = {'interval': '5s'}
        values = [3, 5.2, 13, 4]

        result_regex = """# HELP prometheus_target_interval_length_seconds Actual intervals between scrapes.
# TYPE prometheus_target_interval_length_seconds summary
prometheus_target_interval_length_seconds_count{interval="5s"} 4 \d*(?:.\d*)?
prometheus_target_interval_length_seconds_sum{interval="5s"} 25.2 \d*(?:.\d*)?
prometheus_target_interval_length_seconds{interval="5s",quantile="0.5"} 4.0 \d*(?:.\d*)?
prometheus_target_interval_length_seconds{interval="5s",quantile="0.9"} 5.2 \d*(?:.\d*)?
prometheus_target_interval_length_seconds{interval="5s",quantile="0.99"} 5.2 \d*(?:.\d*)?$"""

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        f = TextFormatter(True)
        result = f.marshall_collector(s)

        self.assertTrue(re.match(result_regex, result))

    def test_registry_marshall(self):

        format_times = 10

        counter_data = (
            ({'c_sample': '1'}, 100),
            ({'c_sample': '2'}, 200),
            ({'c_sample': '3'}, 300),
            ({'c_sample': '1', 'c_subsample': 'b'}, 400),
        )

        gauge_data = (
            ({'g_sample': '1'}, 500),
            ({'g_sample': '2'}, 600),
            ({'g_sample': '3'}, 700),
            ({'g_sample': '1', 'g_subsample': 'b'}, 800),
        )

        summary_data = (
            ({'s_sample': '1'}, range(1000, 2000, 4)),
            ({'s_sample': '2'}, range(2000, 3000, 20)),
            ({'s_sample': '3'}, range(3000, 4000, 13)),
            ({'s_sample': '1', 's_subsample': 'b'}, range(4000, 5000, 47)),
        )

        registry = Registry()
        counter = Counter("counter_test", "A counter.", {'type': "counter"})
        gauge = Gauge("gauge_test", "A gauge.", {'type': "gauge"})
        summary = Summary("summary_test", "A summary.", {'type': "summary"})

        # Add data
        [counter.set(c[0], c[1]) for c in counter_data]
        [gauge.set(g[0], g[1]) for g in gauge_data]
        [summary.add(i[0], s) for i in summary_data for s in i[1]]

        registry.register(counter)
        registry.register(gauge)
        registry.register(summary)

        valid_regex = """# HELP counter_test A counter.
# TYPE counter_test counter
counter_test{c_sample="1",c_subsample="b",type="counter"} 400
counter_test{c_sample="1",type="counter"} 100
counter_test{c_sample="2",type="counter"} 200
counter_test{c_sample="3",type="counter"} 300
# HELP gauge_test A gauge.
# TYPE gauge_test gauge
gauge_test{g_sample="1",g_subsample="b",type="gauge"} 800
gauge_test{g_sample="1",type="gauge"} 500
gauge_test{g_sample="2",type="gauge"} 600
gauge_test{g_sample="3",type="gauge"} 700
# HELP summary_test A summary.
# TYPE summary_test summary
summary_test_count{s_sample="1",s_subsample="b",type="summary"} \d*(?:.\d*)?
summary_test_count{s_sample="1",type="summary"} \d*(?:.\d*)?
summary_test_count{s_sample="2",type="summary"} \d*(?:.\d*)?
summary_test_count{s_sample="3",type="summary"} \d*(?:.\d*)?
summary_test_sum{s_sample="1",s_subsample="b",type="summary"} \d*(?:.\d*)?
summary_test_sum{s_sample="1",type="summary"} \d*(?:.\d*)?
summary_test_sum{s_sample="2",type="summary"} \d*(?:.\d*)?
summary_test_sum{s_sample="3",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.5",s_sample="1",s_subsample="b",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.5",s_sample="1",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.5",s_sample="2",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.5",s_sample="3",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.9",s_sample="1",s_subsample="b",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.9",s_sample="1",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.9",s_sample="2",type="summary"} 2\d*(?:.\d*)?
summary_test{quantile="0.9",s_sample="3",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.99",s_sample="1",s_subsample="b",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.99",s_sample="1",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.99",s_sample="2",type="summary"} \d*(?:.\d*)?
summary_test{quantile="0.99",s_sample="3",type="summary"} \d*(?:.\d*)?
"""
        f = TextFormatter()
        self.maxDiff = None
        # Check multiple times to ensure multiple marshalling requests
        for i in range(format_times):
            self.assertTrue(re.match(valid_regex, f.marshall(registry).decode()))


class TestProtobufFormat(unittest.TestCase):

    # Test Utils
    def _create_protobuf_object(self, data, metrics, metric_type,
                                const_labels={}, ts=False):
        pb_metrics = []
        for i in metrics:
            labels = [pmp.LabelPair(name=k, value=v) for k, v in i[0].items()]
            c_labels = [
                pmp.LabelPair(name=k, value=v) for k, v in const_labels.items()]
            labels.extend(c_labels)

            if metric_type == pmp.COUNTER:
                metric = pmp.Metric(
                    counter=pmp.Counter(value=i[1]),
                    label=labels)
            elif metric_type == pmp.GAUGE:
                metric = pmp.Metric(
                    gauge=pmp.Gauge(value=i[1]),
                    label=labels)
            elif metric_type == pmp.SUMMARY:
                quantiles = []

                for k, v in i[1].items():
                    if not isinstance(k, str):
                        q = pmp.Quantile(quantile=k, value=v)
                        quantiles.append(q)

                metric = pmp.Metric(
                    summary=pmp.Summary(quantile=quantiles,
                                        sample_sum=i[1]['sum'],
                                        sample_count=i[1]['count']),
                    label=labels)
            elif metric_type == pmp.HISTOGRAM:
                buckets = []

                for k, v in i[1].items():
                    if not isinstance(k, str):
                        bucket = pmp.Bucket(
                            cumulative_count=v, upper_bound=k)
                        buckets.append(bucket)

                metric = pmp.Metric(
                    summary=pmp.Histogram(buckets=buckets,
                                          histogram_sum=i[1]['sum'],
                                          histogram_count=i[1]['count']),
                    label=labels)

            else:
                raise TypeError("Not a valid metric")

            if ts:
                metric.timestamp_ms = int(
                    datetime.datetime.now(
                        tz=datetime.timezone.utc).timestamp() * 1000)

            pb_metrics.append(metric)

        valid_result = pmp.MetricFamily(
            name=data['name'],
            help=data['doc'],
            type=metric_type,
            metric=pb_metrics
        )

        return valid_result

    def _protobuf_metric_equal(self, ptb1, ptb2):
        if ptb1 is ptb2:
            return True

        if not ptb1 or not ptb2:
            return False

        # start all the filters
        # 1st level:  Metric Family
        if (ptb1.name != ptb2.name) or\
           (ptb1.help != ptb2.help) or\
           (ptb1.type != ptb2.type) or\
           (len(ptb1.metric) != len(ptb2.metric)):
            return False

        def sort_metric(v):
            """ Small function to order the lists of protobuf """
            x = sorted(v.label, key=lambda x: x.name + x.value)
            return("".join([i.name + i.value for i in x]))

        # Before continuing, sort stuff
        mts1 = sorted(ptb1.metric, key=sort_metric)
        mts2 = sorted(ptb2.metric, key=sort_metric)

        # Now that they are ordered we can compare each element with each
        for k, m1 in enumerate(mts1):
            m2 = mts2[k]

            # Check ts
            if m1.timestamp_ms != m2.timestamp_ms:
                return False

            # Check value
            if ptb1.type == pmp.COUNTER and ptb2.type == pmp.COUNTER:
                if m1.counter.value != m2.counter.value:
                    return False
            elif ptb1.type == pmp.GAUGE and ptb2.type == pmp.GAUGE:
                if m1.gauge.value != m2.gauge.value:
                    return False
            elif ptb1.type == pmp.SUMMARY and ptb2.type == pmp.SUMMARY:
                mm1, mm2 = m1.summary, m2.summary
                if ((mm1.sample_count != mm2.sample_count) or
                        (mm1.sample_sum != mm2.sample_sum)):
                    return False

                # order quantiles to test
                mm1_quantiles = sorted(
                    [(x.quantile, x.value) for x in mm1.quantile])
                mm2_quantiles = sorted(
                    [(x.quantile, x.value) for x in mm2.quantile])

                if mm1_quantiles != mm2_quantiles:
                    return False

            elif ptb1.type == pmp.HISTOGRAM and ptb2.type == pmp.HISTOGRAM:
                mm1, mm2 = m1.summary, m2.summary
                if ((mm1.sample_count != mm2.sample_count) or
                        (mm1.sample_sum != mm2.sample_sum)):
                    return False

                # order buckets to test
                # mm1_buckets = sorted(mm1.bucket, key=lambda x: x.bucket)
                mm1_buckets = sorted(
                    [(x.upper_bound, x.cumulative_count) for x in mm1.bucket])
                # mm2_buckets = sorted(mm2.bucket, key=lambda x: x.bucket)
                mm2_buckets = sorted(
                    [(x.upper_bound, x.cumulative_count) for x in mm2.bucket])

                if mm1_buckets != mm2_buckets:
                    return False

            else:
                return False

            # Check labels
            # Sort labels
            l1 = sorted(m1.label, key=lambda x: x.name + x.value)
            l2 = sorted(m2.label, key=lambda x: x.name + x.value)
            if not all([l.name == l2[k].name and l.value == l2[k].value for k, l in enumerate(l1)]):
                return False

        return True

    def test_create_protobuf_object_wrong(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        values = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
        )

        with self.assertRaises(TypeError) as context:
            self._create_protobuf_object(data, values, 7)

        self.assertEqual("Not a valid metric", str(context.exception))

    def test_test_timestamp(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        values = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
        )

        c = self._create_protobuf_object(data, values, pmp.COUNTER, {})
        for i in c.metric:
            self.assertEqual(0, i.timestamp_ms)

        c = self._create_protobuf_object(data, values, pmp.COUNTER, {}, True)
        for i in c.metric:
            self.assertIsNotNone(i.timestamp_ms)

        self.assertEqual(c, c)
        self.assertTrue(self._protobuf_metric_equal(c, c))
        time.sleep(0.5)
        c2 = self._create_protobuf_object(data, values, pmp.COUNTER, {}, True)
        self.assertFalse(self._protobuf_metric_equal(c, c2))

    def test_test_protobuf_metric_equal_not_metric(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        values = (({"device": "mobile", 'country': "us"}, 654),
                  ({'country': "sp", "device": "desktop"}, 520))
        pt1 = self._create_protobuf_object(data, values, pmp.COUNTER)

        self.assertFalse(self._protobuf_metric_equal(pt1, None))
        self.assertFalse(self._protobuf_metric_equal(None, pt1))

    def test_test_protobuf_metric_equal_not_basic_data(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        pt1 = self._create_protobuf_object(data, (), pmp.COUNTER)

        data2 = data.copy()
        data2['name'] = "other"
        pt2 = self._create_protobuf_object(data2, (), pmp.COUNTER)
        self.assertFalse(self._protobuf_metric_equal(pt1, pt2))

        data2 = data.copy()
        data2['doc'] = "other"
        pt2 = self._create_protobuf_object(data2, (), pmp.COUNTER)
        self.assertFalse(self._protobuf_metric_equal(pt1, pt2))

        pt2 = self._create_protobuf_object(data, (), pmp.SUMMARY)
        self.assertFalse(self._protobuf_metric_equal(pt1, pt2))

        pt3 = self._create_protobuf_object(data, (), pmp.HISTOGRAM)
        self.assertFalse(self._protobuf_metric_equal(pt2, pt3))

    def test_test_protobuf_metric_equal_not_labels(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        values = (({"device": "mobile", 'country': "us"}, 654),)
        pt1 = self._create_protobuf_object(data, values, pmp.COUNTER)

        values2 = (({"device": "mobile", 'country': "es"}, 654),)
        pt2 = self._create_protobuf_object(data, values2, pmp.COUNTER)

        self.assertFalse(self._protobuf_metric_equal(pt1, pt2))

    def test_test_protobuf_metric_equal_counter(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        counter_data = (
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'pt2': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 521),
                        ({'country': "us", "device": "mobile"}, 654),),
                'pt2': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': False
            },
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 520),
                        ({"device": "mobile", 'country': "us"}, 654),),
                'pt2': (({"device": "desktop", 'country': "sp"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
            {
                'pt1': (({"device": "mobile", 'country': "us"}, 654),
                        ({'country': "sp", "device": "desktop"}, 520)),
                'pt2': (({"device": "desktop", 'country': "sp"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
        )

        for i in counter_data:
            p1 = self._create_protobuf_object(data, i['pt1'], pmp.COUNTER)
            p2 = self._create_protobuf_object(data, i['pt2'], pmp.COUNTER)

            if i['ok']:
                self.assertTrue(self._protobuf_metric_equal(p1, p2))
            else:
                self.assertFalse(self._protobuf_metric_equal(p1, p2))

    def test_test_protobuf_metric_equal_gauge(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        gauge_data = (
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'pt2': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 521),
                        ({'country': "us", "device": "mobile"}, 654),),
                'pt2': (({'country': "sp", "device": "desktop"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': False
            },
            {
                'pt1': (({'country': "sp", "device": "desktop"}, 520),
                        ({"device": "mobile", 'country': "us"}, 654),),
                'pt2': (({"device": "desktop", 'country': "sp"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
            {
                'pt1': (({"device": "mobile", 'country': "us"}, 654),
                        ({'country': "sp", "device": "desktop"}, 520)),
                'pt2': (({"device": "desktop", 'country': "sp"}, 520),
                        ({'country': "us", "device": "mobile"}, 654),),
                'ok': True
            },
        )

        for i in gauge_data:
            p1 = self._create_protobuf_object(data, i['pt1'], pmp.GAUGE)
            p2 = self._create_protobuf_object(data, i['pt2'], pmp.GAUGE)

            if i['ok']:
                self.assertTrue(self._protobuf_metric_equal(p1, p2))
            else:
                self.assertFalse(self._protobuf_metric_equal(p1, p2))

    def test_test_protobuf_metric_equal_summary(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }

        gauge_data = (
            {
                'pt1': (({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
                        ({'interval': "10s"}, {0.5: 90, 0.9: 149, 0.99: 150, "sum": 385, "count": 10}),),
                'pt2': (({'interval': "10s"}, {0.5: 90, 0.9: 149, 0.99: 150, "sum": 385, "count": 10}),
                        ({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4})),
                'ok': True
            },
            {
                'pt1': (({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
                        ({'interval': "10s"}, {0.5: 90, 0.9: 149, 0.99: 150, "sum": 385, "count": 10}),),
                'pt2': (({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
                        ({'interval': "10s"}, {0.5: 90, 0.9: 150, 0.99: 150, "sum": 385, "count": 10}),),
                'ok': False
            },
            {
                'pt1': (({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
                        ({'interval': "10s"}, {0.5: 90, 0.9: 149, 0.99: 150, "sum": 385, "count": 10}),),
                'pt2': (({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
                        ({'interval': "10s"}, {0.5: 90, 0.9: 149, 0.99: 150, "sum": 385, "count": 11}),),
                'ok': False
            },
        )

        for i in gauge_data:
            p1 = self._create_protobuf_object(data, i['pt1'], pmp.SUMMARY)
            p2 = self._create_protobuf_object(data, i['pt2'], pmp.SUMMARY)

            if i['ok']:
                self.assertTrue(self._protobuf_metric_equal(p1, p2))
            else:
                self.assertFalse(self._protobuf_metric_equal(p1, p2))

#     # Finish Test Utils
#     # ######################################

    def test_headers(self):
        f = BinaryFormatter()
        result = {
            'Content-Type': 'application/vnd.google.protobuf; proto=io.prometheus.client.MetricFamily; encoding=delimited'
        }

        self.assertEqual(result, f.get_headers())

    def test_wrong_format(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

        f = BinaryFormatter()

        c = Collector(**data)

        with self.assertRaises(TypeError) as context:
            f.marshall_collector(c)

        self.assertEqual('Not a valid object format', str(context.exception))

    def test_counter_format(self):

        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }
        c = Counter(**data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
        )

        # Construct the result to compare
        valid_result = self._create_protobuf_object(
            data, counter_data, pmp.COUNTER)

        # Add data to the collector
        for i in counter_data:
            c.set_value(i[0], i[1])

        f = BinaryFormatter()

        result = f.marshall_collector(c)

        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_counter_format_with_const_labels(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }
        c = Counter(**data)

        counter_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
        )

        # Construct the result to compare
        valid_result = self._create_protobuf_object(
            data, counter_data, pmp.COUNTER, data['const_labels'])

        # Add data to the collector
        for i in counter_data:
            c.set_value(i[0], i[1])

        f = BinaryFormatter()

        result = f.marshall_collector(c)

        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_gauge_format(self):

        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': None,
        }
        g = Gauge(**data)

        gauge_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
        )

        # Construct the result to compare
        valid_result = self._create_protobuf_object(
            data, gauge_data, pmp.GAUGE)

        # Add data to the collector
        for i in gauge_data:
            g.set_value(i[0], i[1])

        f = BinaryFormatter()

        result = f.marshall_collector(g)

        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_gauge_format_with_const_labels(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }
        g = Gauge(**data)

        gauge_data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
        )

        # Construct the result to compare
        valid_result = self._create_protobuf_object(
            data, gauge_data, pmp.GAUGE, data['const_labels'])

        # Add data to the collector
        for i in gauge_data:
            g.set_value(i[0], i[1])

        f = BinaryFormatter()

        result = f.marshall_collector(g)

        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_one_summary_format(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }

        labels = {'handler': '/static'}
        values = [3, 5.2, 13, 4]

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        tmp_valid_data = [
            (labels, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
        ]
        valid_result = self._create_protobuf_object(data, tmp_valid_data,
                                                    pmp.SUMMARY)

        f = BinaryFormatter()

        result = f.marshall_collector(s)
        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_one_summary_format_with_const_labels(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

        labels = {'handler': '/static'}
        values = [3, 5.2, 13, 4]

        s = Summary(**data)

        for i in values:
            s.add(labels, i)

        tmp_valid_data = [
            (labels, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
        ]
        valid_result = self._create_protobuf_object(data, tmp_valid_data,
                                                    pmp.SUMMARY,
                                                    data['const_labels'])

        f = BinaryFormatter()

        result = f.marshall_collector(s)
        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_summary_format(self):
        data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {},
        }

        summary_data = (
            ({'interval': "5s"}, [3, 5.2, 13, 4]),
            ({'interval': "10s"}, [1.3, 1.2, 32.1, 59.2, 109.46, 70.9]),
            ({'interval': "10s", 'method': "fast"}, [5, 9.8, 31, 9.7, 101.4]),
        )

        s = Summary(**data)

        for i in summary_data:
            for j in i[1]:
                s.add(i[0], j)

        tmp_valid_data = [
            ({'interval': "5s"}, {0.5: 4.0, 0.9: 5.2, 0.99: 5.2, "sum": 25.2, "count": 4}),
            ({'interval': "10s"}, {0.5: 32.1, 0.9: 59.2, 0.99: 59.2, "sum": 274.15999999999997, "count": 6}),
            ({'interval': "10s", 'method': "fast"}, {0.5: 9.7, 0.9: 9.8, 0.99: 9.8, "sum": 156.9, "count": 5}),
        ]
        valid_result = self._create_protobuf_object(data, tmp_valid_data,
                                                    pmp.SUMMARY)

        f = BinaryFormatter()

        result = f.marshall_collector(s)
        self.assertTrue(self._protobuf_metric_equal(valid_result, result))

    def test_registry_marshall_counter(self):

        format_times = 10

        counter_data = (
            ({'c_sample': '1', 'c_subsample': 'b'}, 400),
        )

        registry = Registry()
        counter = Counter("counter_test", "A counter.", {'type': "counter"})

        # Add data
        [counter.set(c[0], c[1]) for c in counter_data]

        registry.register(counter)

        valid_result = (b'[\n\x0ccounter_test\x12\nA counter.\x18\x00"=\n\r'
                        b'\n\x08c_sample\x12\x011\n\x10\n\x0bc_subsample\x12'
                        b'\x01b\n\x0f\n\x04type\x12\x07counter\x1a\t\t\x00\x00'
                        b'\x00\x00\x00\x00y@')
        f = BinaryFormatter()

        # Check multiple times to ensure multiple marshalling requests
        for i in range(format_times):
            self.assertEqual(valid_result, f.marshall(registry))

    def test_registry_marshall_gauge(self):
        format_times = 10

        gauge_data = (
            ({'g_sample': '1', 'g_subsample': 'b'}, 800),
        )

        registry = Registry()
        gauge = Gauge("gauge_test", "A gauge.", {'type': "gauge"})

        # Add data
        [gauge.set(g[0], g[1]) for g in gauge_data]

        registry.register(gauge)

        valid_result = (b'U\n\ngauge_test\x12\x08A gauge.\x18\x01";'
                        b'\n\r\n\x08g_sample\x12\x011\n\x10\n\x0bg_subsample'
                        b'\x12\x01b\n\r\n\x04type\x12\x05gauge\x12\t\t\x00'
                        b'\x00\x00\x00\x00\x00\x89@')

        f = BinaryFormatter()

        # Check multiple times to ensure multiple marshalling requests
        for i in range(format_times):
            self.assertEqual(valid_result, f.marshall(registry))

    def test_registry_marshall_summary(self):
        format_times = 10

        summary_data = (
            ({'s_sample': '1', 's_subsample': 'b'}, range(4000, 5000, 47)),
        )

        registry = Registry()
        summary = Summary("summary_test", "A summary.", {'type': "summary"})

        # Add data
        [summary.add(i[0], s) for i in summary_data for s in i[1]]

        registry.register(summary)

        valid_result = (b'\x99\x01\n\x0csummary_test\x12\nA summary.'
                        b'\x18\x02"{\n\r\n\x08s_sample\x12\x011\n\x10\n'
                        b'\x0bs_subsample\x12\x01b\n\x0f\n\x04type\x12\x07'
                        b'summary"G\x08\x16\x11\x00\x00\x00\x00\x90"\xf8@'
                        b'\x1a\x12\t\x00\x00\x00\x00\x00\x00\xe0?\x11\x00'
                        b'\x00\x00\x00\x00\x8b\xb0@\x1a\x12\t\xcd\xcc\xcc'
                        b'\xcc\xcc\xcc\xec?\x11\x00\x00\x00\x00\x00v\xb1@'
                        b'\x1a\x12\t\xaeG\xe1z\x14\xae\xef?\x11\x00\x00\x00'
                        b'\x00\x00\xa5\xb1@')

        f = BinaryFormatter()

        # Check multiple times to ensure multiple marshalling requests
        for i in range(format_times):
            self.assertEqual(valid_result, f.marshall(registry))




import logging

from .formats import BinaryFormatter, TextFormatter
from typing import Callable, Set, Union

logger = logging.getLogger(__name__)

# type aliases
BinaryFormatterClass = Callable[[bool], BinaryFormatter]
TextFormatterClass = Callable[[bool], TextFormatter]
FormatterType = Union[BinaryFormatterClass, TextFormatterClass]


ProtobufAccepts = set(
    ['application/vnd.google.protobuf',
     'proto=io.prometheus.client.MetricFamily',
     'encoding=delimited'])

TextAccepts = set(['text/plain', 'version=0.0.4'])


def negotiate(accepts: Set[str]) -> FormatterType:
    ''' Negotiate a response format by scanning through the ACCEPTS
    header and selecting the most efficient format.

    :param accepts: a set of ACCEPT headers fields extracted from a request.

    :returns: a formatter class that should be used to form up the
      response into the appropriate representation.
    '''
    if not isinstance(accepts, set):
        raise TypeError(
            'Expected a set but got {}'.format(type(accepts)))

    formatter = TextFormatter  # type: FormatterType

    if ProtobufAccepts.issubset(accepts):
        formatter = BinaryFormatter
    elif TextAccepts.issubset(accepts):
        formatter = TextFormatter

    logger.debug(
        'negotiating {} resulted in choosing {}'.format(
            accepts, formatter.__name__))
    return formatter




import unittest

from aioprometheus.formats import BinaryFormatter, TextFormatter
from aioprometheus.negotiator import negotiate


class TestNegotiate(unittest.TestCase):

    def test_protobuffer(self):
        ''' check that a protobuf formatter is returned '''
        headers = (
            "proto=io.prometheus.client.MetricFamily;application/vnd.google.protobuf;encoding=delimited",
            "application/vnd.google.protobuf;proto=io.prometheus.client.MetricFamily;encoding=delimited",
            "encoding=delimited;application/vnd.google.protobuf;proto=io.prometheus.client.MetricFamily")

        for accept in headers:
            self.assertEqual(
                BinaryFormatter, negotiate(set(accept.split(';'))))

    def test_text_004(self):
        ''' check that a text formatter is returned for version 0.0.4 '''
        headers = (
            "text/plain; version=0.0.4",
            "text/plain;version=0.0.4",
            "version=0.0.4; text/plain")

        for accept in headers:
            self.assertEqual(
                TextFormatter, negotiate(set(accept.split(';'))))

    def test_text_default(self):
        ''' check that a text formatter is returned for plain text '''
        headers = (
            "text/plain;",)

        for accept in headers:
            self.assertEqual(
                TextFormatter, negotiate(set(accept.split(';'))))

    def test_default(self):
        ''' check that a text formatter is returned if no matches '''
        headers = (
            "application/json",
            "*/*",
            "application/nothing")

        for accept in headers:
            self.assertEqual(
                TextFormatter, negotiate(set(accept.split(';'))))




from .collectors import Collector, Counter, Gauge, Histogram, Summary
from typing import List, Union

CollectorsType = Union[Counter, Gauge, Histogram, Summary]


class CollectorRegistry(object):
    ''' This class implements the metrics collector registry.

    Collectors in the registry must comply with the Collector interface
    which means that they inherit from the base Collector object and implement
    a no-argument method called 'get_all' that returns a list of Metric
    objects.
    '''

    def __init__(self) -> None:
        self.collectors = {}  # type: Dict[str, CollectorsType]

    def register(self, collector: CollectorsType) -> None:
        ''' Register a collector

        :raises: TypeError if collector is not an instance of
          :class:`Collector`.
        :raises: ValueError if collector is already registered.
        '''
        if not isinstance(collector, Collector):
            raise TypeError(
                'Invalid collector type: {}'.format(collector))

        if collector.name in self.collectors:
            raise ValueError(
                "Collector {} is already registered".format(
                    collector.name))

        self.collectors[collector.name] = collector

    def deregister(self, name: str) -> None:
        ''' Deregister a collector.

        :param name: The name of the collector to deregister.
        '''
        del self.collectors[name]

    def get(self, name: str) -> CollectorsType:
        ''' Get a collector

        :param name: The name of the collector to fetch.

        :raises: KeyError if collector is not found.
        '''
        return self.collectors[name]

    def get_all(self) -> List[CollectorsType]:
        ''' Return a list of all collectors '''
        return list(self.collectors.values())


# The default registry.
Registry = CollectorRegistry



'''
This module implements the abstract service discovery API interface.
'''

import abc


class IDiscoveryAgent(abc.ABC):
    '''
    This class represents the abstract base class for a discovery agent.
    '''

    async def register(self, metrics_server):
        '''
        Register a Prometheus metrics server with service discovery.

        :param metrics_server: an instance of a :class:`MetricsServer`
        '''
        raise NotImplementedError

    async def deregister(self, metrics_server):
        '''
        Register a Prometheus metrics server from service discovery.

        :param metrics_server: an instance of a :class:`MetricsServer`
        '''
        raise NotImplementedError




import asyncio

import aiohttp
import aiohttp.web

from urllib.parse import urljoin
from .formats import BinaryFormatter

# imports only used for type annotations
from asyncio.base_events import BaseEventLoop
from .registry import CollectorRegistry


class Pusher(object):
    '''
    This class can be used in applications that can't support the
    standard pull strategy. The pusher object pushers the metrics in a
    registry to a pushgateway which can be scraped by Prometheus.
    '''

    PATH = "/metrics/job/{0}"

    def __init__(self,
                 job_name: str,
                 addr: str,
                 loop: BaseEventLoop = None) -> None:
        '''

        :param job_name: The name of the job.

        :param addr: The address of the push gateway. The default port the
          push gateway listens on is 9091 so the address will typically be
          something like this ``http://hostname:9091``.

        :param loop: The event loop instance to use. If no loop is specified
          then the default event loop will be used.
        '''
        self.job_name = job_name
        self.addr = addr
        self.loop = loop or asyncio.get_event_loop()
        self.formatter = BinaryFormatter()
        self.headers = self.formatter.get_headers()
        self.path = urljoin(self.addr, self.PATH.format(job_name))

    async def add(self,
                  registry: CollectorRegistry) -> aiohttp.web.Response:
        '''
        Add works like replace, but only metrics with the same name as the
        newly pushed metrics are replaced.
        '''
        with aiohttp.ClientSession(loop=self.loop) as session:
            payload = self.formatter.marshall(registry)
            resp = await session.post(
                self.path, data=payload, headers=self.headers)
        await resp.release()
        return resp

    async def replace(self,
                      registry: CollectorRegistry) -> aiohttp.web.Response:
        '''
        ``replace`` pushes new values for a group of metrics to the push
        gateway.

        .. note::

            All existing metrics with the same grouping key specified in the
            URL will be replaced with the new metrics value.

        '''
        with aiohttp.ClientSession(loop=self.loop) as session:
            payload = self.formatter.marshall(registry)
            resp = await session.put(
                self.path, data=payload, headers=self.headers)
        await resp.release()
        return resp

    async def delete(self,
                     registry: CollectorRegistry) -> aiohttp.web.Response:
        '''
        ``delete`` deletes metrics from the push gateway. All metrics with
        the grouping key specified in the URL are deleted.
        '''
        with aiohttp.ClientSession(loop=self.loop) as session:
            payload = self.formatter.marshall(registry)
            resp = await session.delete(
                self.path, data=payload, headers=self.headers)
        await resp.release()
        return resp



'''
This module is built upon the :mod:`unittest` standard library module.
It simplifies event loop related boilerplate code when testing code built
upon the :mod:`asyncio` standard library module.

This module imports the :mod:`unittest` package, overrides some of
its features and adds new ones.

Enhanced :class:`unittest.TestCase`:

- a new event loop is created at the beginning of each test. This event loop
  is set as the default event loop. The event loop is stopped, closed and
  destroyed at the end of each test.

- The :meth:`~TestCase.setUp()` and :meth:`~TestCase.tearDown()` methods can
  be coroutine functions.

- a test method in a TestCase (e.g. `test_*`) declared as a coroutine function
  or returning a coroutine will run in the event loop.

- cleanup functions registered with :meth:`~TestCase.addCleanup()` can be
  coroutine functions.


'''

import asyncio
import functools
import gc
import inspect
import sys
import unittest

if False:
    from asyncio.base_events import BaseEventLoop
    from asyncio.events import BaseDefaultEventLoopPolicy

PY34 = sys.version_info >= (3, 4)
PY35 = sys.version_info >= (3, 5)


def isawaitable(obj):
    ''' Return True if the object is an awaitable or is a function that
    returns an awaitable.

    This function is used internally by aiotesting.
    '''
    if PY35:
        result = inspect.iscoroutinefunction(obj) or inspect.isawaitable(obj)

    elif PY34:
        result = (isinstance(obj, asyncio.Future) or
                  asyncio.iscoroutine(obj) or
                  hasattr(obj, '__await__'))
    else:
        raise Exception(
            'isawaitable is not supported on Python {}'.format(
                sys.version_info))
    return result


class AsyncioTestCase(unittest.TestCase):
    '''
    This class enhances :class:`unittest.TestCase` to work nicely with
    tests that use the :mod:`asyncio` module.

    By default a new event loop is created for each test and is set as the
    default event loop. The setUp and tearDown methods can also be defined
    as coroutines.

    The behaviour of this class can be configured by defining some class
    attributes.

    There may be times when you need to run a specific event loop. This can
    be done by specifying a loop policy as a class attribute.

    .. code-block:: python

        from .my_code import MyEventLoopPolicy

        class MyEventLoop_TestCase(AsyncioTestCase):

            loop_policy = MyEventLoopPolicy()

    In a number of different situations it can be useful to have tests
    time-out after some reasonable period of time. Tests can sometimes
    stall the test suite progression by effectively hanging in the event
    loop while waiting for a future that never fires. For example, certain
    kinds of tests that rely on external processes or services and during
    early development of asynchronous test cases,

    With AsyncioTestCase you can use the ``@TestTimeout`` decorator on
    specific tests.

    .. code-block:: python

        class MyAsyncTestCase(AsyncioTestCase):

            @TestTimeout(3.0)
            def test_func(self):
                yield from asyncio.sleep(2.0)

    If :attr:`~aiotesting.AsyncioTestCase` is set to True then a class-wide
    timeout is used for each test in the TestCase. Alternatively, a class
    decorator can be used.

    .. code-block:: python

        @TestCaseTimeout(3.0)
        class MyAsyncTestCase(AsyncioTestCase):

            async def test_func(self):
                await asyncio.sleep(2.0)

    '''

    # A specific event loop can be used by defining an event loop policy.
    loop_policy = None  # type: BaseDefaultEventLoopPolicy

    loop = None  # type: BaseEventLoop

    # To stop poorly written tests from blocking the efficient progress
    # of running test cases, a default test timeout can be set. If you
    # need timeouts for specific tests then use the `@TestTimeout`
    # decorator..
    timeout = None  # type: float

    def run(self, result=None):
        '''
        Overrides the :meth:`~TestCase.run()` method so that
        :meth:`~TestCase.setUp()` and :meth:`~TestCase.tearDown()` can be
        coroutines.
        '''
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, 'startTestRun', None)
            if startTestRun is not None:
                startTestRun()

        result.startTest(self)

        testMethod = getattr(self, self._testMethodName)
        if (getattr(self.__class__, "__unittest_skip__", False) or
                getattr(testMethod, "__unittest_skip__", False)):
            # If the class or method was skipped.
            try:
                skip_why = (
                    getattr(self.__class__, '__unittest_skip_why__', '') or
                    getattr(testMethod, '__unittest_skip_why__', ''))
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return
        expecting_failure = getattr(testMethod,
                                    "__unittest_expecting_failure__", False)
        outcome = unittest.case._Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self):
                self._setUp()
            if outcome.success:
                outcome.expecting_failure = expecting_failure
                with outcome.testPartExecutor(self, isTest=True):
                    self._run_test_method(testMethod)
                outcome.expecting_failure = False
                with outcome.testPartExecutor(self):
                    self._tearDown()

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(
                            result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, 'stopTestRun', None)
                if stopTestRun is not None:
                    stopTestRun()

            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome ->
            # outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None

            # clear the outcome, no more needed
            self._outcome = None

    def addCleanup(self, function, *args, **kwargs):
        '''
        Add a function, with arguments, to be called when the test is
        completed. If function is a coroutine function, it will be run by the
        event loop before it is destroyed.
        '''
        if asyncio.iscoroutinefunction(function):
            return super().addCleanup(self.loop.run_until_complete,
                                      function(*args, **kwargs))

        return super().addCleanup(function, *args, **kwargs)

    def debug(self):
        '''Run the test without collecting errors in a TestResult'''
        self._setUp()
        meth = getattr(self, self._testMethodName)
        self._run_test_method(meth())
        self._tearDown()
        while self._cleanups:
            function, args, kwargs = self._cleanups.pop(-1)
            function(*args, **kwargs)

    # internals

    def _setUp(self):
        ''' Create a new loop for each test case '''
        asyncio.set_event_loop_policy(self.loop_policy)
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        if asyncio.iscoroutinefunction(self.setUp):
            self.loop.run_until_complete(self.setUp())
        else:
            self.setUp()

    def _tearDown(self):
        ''' Destroy the event loop '''
        if asyncio.iscoroutinefunction(self.tearDown):
            self.loop.run_until_complete(self.tearDown())
        else:
            self.tearDown()

        if not isinstance(self.loop, asyncio.AbstractEventLoop):
            raise Exception('Invalid event loop: ', self.loop)
        if self.loop.is_running():
            self.loop.stop()
        self.loop.close()
        del self.loop
        asyncio.set_event_loop_policy(None)
        asyncio.set_event_loop(None)

        # By explicitly forcing a garbage collection here,
        # the event loop will report any remaining sockets
        # and coroutines left in the event loop which indicates
        # that further cleanup actions should be implemented
        # in the code under test.
        gc.collect()

    def _run_test_method(self, method):
        ''' Run a test method.

        If running the method returns an awaitable object then run it in the
        event loop.
        '''
        result = method()
        if isawaitable(result):
            if self.timeout:
                result = asyncio.wait_for(
                    result, timeout=self.timeout, loop=self.loop)
            self.loop.run_until_complete(result)


def TestCaseTimeout(timeout):
    '''
    This decorator function applies a default test timeout to the wrapped
    class. When the wrapped class is an AsyncTestCase class then the timeout
    attribute added to the class is used as the default timeout for all test
    methods within the TestCase.

    .. code-block:: python

        from aiotesting import AsyncioTestCase, TestCaseTimeout

        @TestCaseTimeout(1.0)
        class MyTestCase(AsyncioTestCase):

            async def test_func(self):
                await example_func()

    '''
    def wrapper(cls):
        cls.timeout = timeout
        return cls

    return wrapper


def TestTimeout(timeout):
    '''
    This decorator function applies a test timeout to a test method on a test
    by test basis.

    .. code-block:: python

        from aiotesting import AsyncioTestCase, TestTimeout

        class MyTestCase(AsyncioTestCase):

            @TestTimeout(1.0)
            async def test_func(self):
                await example_func()

    '''
    def decorating_function(user_function):

        def wrapper(*args, **kwargs):

            result = user_function(*args, **kwargs)

            if not isawaitable(result):
                raise Exception(
                    'Invalid use of TestTimeout decorator. Expected an '
                    'awaitable but got {}, type={}'.format(
                        result, type(result)))

            result = asyncio.wait_for(
                result, timeout=timeout, loop=asyncio.get_event_loop())

            return result

        return functools.update_wrapper(wrapper, user_function)

    return decorating_function




# copied from prometheus-python

import unittest

from aioprometheus.metricdict import MetricDict


class TestMetricDict(unittest.TestCase):

    def setUp(self):
        pass

    def test_bad_keys(self):
        with self.assertRaises(TypeError) as context:
            metrics = MetricDict()
            metrics['not_valid'] = "value"

        self.assertEqual('Only accepts dicts as keys', str(context.exception))

    def test_set(self):
        metrics = MetricDict()
        data = (
            ({'a': 1}, 1000),
            ({'b': 2, 'c': 3}, 2000),
            ({'d': 4, 'e': 5, 'f': 6}, 3000),
        )

        for i in data:
            metrics[i[0]] = i[1]

        self.assertEqual(len(data), len(metrics))

    def test_get(self):
        metrics = MetricDict()
        data = (
            ({'a': 1}, 1000),
            ({'b': 2, 'c': 3}, 2000),
            ({'d': 4, 'e': 5, 'f': 6}, 3000),
        )

        for i in data:
            metrics[i[0]] = i[1]

        for i in data:
            self.assertEqual(i[1], metrics[i[0]])

    def test_override(self):
        metrics = MetricDict()
        key = {'a': 1}

        for i in range(100):
            metrics[key] = i
            self.assertEqual(i, metrics[key])

        self.assertEqual(1, len(metrics))

    def test_similar(self):
        metrics = MetricDict()
        data = (
            ({'d': 4, 'e': 5, 'f': 6}, 3000),
            ({'e': 5, 'd': 4, 'f': 6}, 4000),
            ({'d': 4, 'f': 6, 'e': 5}, 5000),
        )

        for i in data:
            metrics[i[0]] = i[1]

        self.assertEqual(1, len(metrics))

    def test_access_by_str(self):
        label = {'b': 2, 'c': 3, 'a': 1}
        access_key = '{"a": 1, "b": 2, "c": 3}'
        bad_access_key = '{"b": 2, "c": 3, "a": 1}'
        value = 100

        metrics = MetricDict()
        metrics[label] = 100

        # Wrong string
        with self.assertRaises(TypeError) as context:
            metrics['dasdasd']
        self.assertEqual('Only accepts dicts as keys', str(context.exception))

        # Access ok with string
        self.assertEqual(value, metrics[access_key])

        # Access ok but wrong key by order
        with self.assertRaises(KeyError) as context:
            metrics[bad_access_key]
        self.assertEqual("'{0}'".format(bad_access_key),
                         str(context.exception))

    def test_empty_key(self):
        metrics = MetricDict()
        iterations = 100

        for i in range(iterations):
            metrics[None] = i
            self.assertEqual(metrics[None], i)

        # Last check (different empty)
        self.assertEqual(iterations - 1, metrics[""])
        self.assertEqual(iterations - 1, metrics[{}])

    def test_delete(self):
        metrics = MetricDict()
        data = (
            ({'d': 4, 'e': 5, 'f': 6}, 3000),
            ({'e': 5, 'd': 4, 'f': 6}, 4000),
            ({'d': 4, 'f': 6, 'e': 5}, 5000),
            ({'d': 41, 'f': 61, 'e': 51}, 6000),
            ({'d': 41, 'e': 51, 'f': 61}, 7000),
            ({'f': 61, 'e': 51, 'd': 41}, 8000),
        )

        for i in data:
            metrics[i[0]] = i[1]

        del metrics[i[0]]

        self.assertEqual(1, len(metrics))

    def test_all(self):
        metrics = MetricDict()
        data = (
            ({'d': 4, 'e': 5, 'f': 6}, 3000),
            ({'e': 5, 'd': 4, 'f': 6}, 4000),
            ({'d': 4, 'f': 6, 'e': 5}, 5000),
            ({'d': 41, 'f': 61, 'e': 51}, 6000),
            ({'d': 41, 'e': 51, 'f': 61}, 7000),
            ({'f': 61, 'e': 51, 'd': 41}, 8000),
        )

        for i in data:
            metrics[i[0]] = i[1]

        self.assertEqual(2, len(metrics))

        self.assertEqual(5000, metrics[{'d': 4, 'e': 5, 'f': 6}])
        self.assertEqual(8000, metrics[{'d': 41, 'f': 61, 'e': 51}])



'''
This module implements part of a custom service discovery approach that uses
``etcd``, a highly consistent distributed key-value store, to store the
metrics service information. This information can be retrieved (using a
separate ``etcd`` reader tool) and made available to the Prometheus server.

Application's instrumented with the ``aioprometheus`` metrics service expose
an endpoint for the Prometheus Server to scrape metrics from. The
``EtcdAgent`` in this module is used by an application to register (and
deregister) its metrics service (i.e. ``/metrics``) endpoint with the
service discovery store (i.e. etcd) using a pre-defined key schema:

.. code-block:: console

    /services/<service_name> = <service_data>

The data contained in the ``service_data`` value is a dict of information
associated with the service:

.. code-block:: json

    {
     "service_name": "service_a",
     "tags": [("hostname", "hostname_a")],
     "address": "http://<addr>:<port>/metrics"
    }

'''

import asyncio
import json
import logging

from aio_etcd.client import Client as EtcdClient
from .agent import IDiscoveryAgent

# imports only used for type annotations
from asyncio.base_events import BaseEventLoop
from ..service import Service


logger = logging.getLogger(__name__)


class EtcdAgent(IDiscoveryAgent):
    '''
    This class implements a service discovery mechansim that allows an
    Prometheus service to register itself with a service discovery tool
    so that Prometheus can discover new applications to monitor. This
    implementation uses etcd as the backing store for service discovery.

    This mechanism stores key values as JSON objects.
    '''

    def __init__(self,
                 service_name: str,
                 tags=(),
                 loop: BaseEventLoop = None,
                 **kwargs) -> None:
        '''

        :param service_name: a string that uniquely defines the service
          name. This name will be used to register the Prometheus metrics
          service for an application in to etcd. The responsibility for
          coming up with unique service names is outside the scope of this
          client.

        :param tags: a optional sequence of key-value tuples.

        :keywords: are passed straight down to the etcd client.
        '''
        self.service_name = service_name
        self.tags = tags
        self.loop = loop or asyncio.get_event_loop()
        self.client = EtcdClient(loop=self.loop, **kwargs)

    async def register(self, metrics_server: Service) -> None:
        '''
        Register a Prometheus metrics server with etcd.

        :param metrics_server: an instance of a :class:`Service`

        :param loop: an event loop instance. If no event loop is provided
          then the default event will be used.
        '''
        key = '/services/{}'.format(self.service_name)
        value = dict(
            service_name=self.service_name,
            tags=self.tags,
            address=metrics_server.url)
        await self.client.write(key, json.dumps(value))

    async def deregister(self, metrics_server: Service) -> None:
        '''
        Register a Prometheus metrics server from etcd.

        :param metrics_server: an instance of a :class:`MetricsServer`

        :param loop: an event loop instance. If no event loop is provided
          then the default event will be used.
        '''
        key = '/services/{}'.format(self.service_name)
        await self.client.delete(key)

    async def close(self) -> None:
        '''
        Stop the agent. This method does not deregister the service.
        '''
        if self.client:
            self.client.close()



''' This module implements a Prometheus metrics text formatter '''

import collections

from .base import IFormatter
from ..collectors import Counter, Gauge, Summary, Histogram
from typing import cast, Callable, Dict, List, Tuple, Union

# imports only used for type annotations
from ..registry import CollectorRegistry

# typing aliases
LabelsType = Dict[str, str]
NumericValueType = Union[int, float]
# ValueType = Union[str, NumericValueType]
SummaryDictKeyType = Union[float, str]  # e.g. sum, 0.25, etc
SummaryDictType = Dict[SummaryDictKeyType, NumericValueType]
HistogramDictKeyType = Union[float, str]  # e.g. sum, 0.25, etc
HistogramDictType = Dict[HistogramDictKeyType, NumericValueType]
CollectorsType = Union[Counter, Gauge, Histogram, Summary]
MetricValueType = Union[NumericValueType, SummaryDictType, HistogramDictType]
MetricTupleType = Tuple[LabelsType, MetricValueType]
FormatterFuncType = Callable[[MetricTupleType, str, LabelsType], List[str]]


HELP_FMT = "# HELP {name} {doc}"
TYPE_FMT = "# TYPE {name} {kind}"
COMMENT_FMT = "# {comment}"
LABEL_FMT = "{key}=\"{value}\""
LABEL_SEPARATOR_FMT = ","
LINE_SEPARATOR_FMT = "\n"
METRIC_FMT = "{name}{labels} {value} {timestamp}"
POS_INF = float("inf")
NEG_INF = float("-inf")


class TextFormatter(IFormatter):
    ''' This formatter encodes into the Protocol Buffers binary format '''

    def __init__(self, timestamp: bool = False) -> None:
        ''' Initialise the text formatter.

        timestamp is a boolean, if you want timestamp in each metric.
        '''
        self.timestamp = timestamp
        self._headers = {
            'Content-Type': 'text/plain; version=0.0.4; charset=utf-8'}

    def get_headers(self) -> Dict[str, str]:
        return self._headers

    def _format_line(self,
                     name: str,
                     labels: LabelsType,
                     value: NumericValueType,
                     const_labels: LabelsType = None) -> str:

        labels = self._unify_labels(labels, const_labels, True)

        labels_str = ""  # type: str
        if labels:
            _labels = [
                LABEL_FMT.format(key=k, value=v)
                for k, v in labels.items()]
            labels_str = LABEL_SEPARATOR_FMT.join(_labels)
            labels_str = "{{{labels}}}".format(labels=labels_str)

        ts = ""  # type: Union[str, int]
        if self.timestamp:
            ts = self._get_timestamp()

        result = METRIC_FMT.format(
            name=name, labels=labels_str, value=value, timestamp=ts)

        return result.strip()

    def _format_counter(self,
                        counter: MetricTupleType,
                        name: str,
                        const_labels: LabelsType) -> List[str]:
        '''
        :param counter: a 2-tuple containing labels and the counter value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        labels, value = counter
        value = cast(NumericValueType, value)  # typing check, no runtime behaviour.
        line = self._format_line(name, labels, value, const_labels)
        return [line]

    def _format_gauge(self,
                      gauge: MetricTupleType,
                      name: str,
                      const_labels: LabelsType) -> List[str]:
        '''
        :param gauge: a 2-tuple containing labels and the gauge value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        labels, value = gauge
        value = cast(NumericValueType, value)  # typing check, no runtime behaviour.
        line = self._format_line(name, labels, value, const_labels)
        return [line]

    def _format_summary(self,
                        summary: MetricTupleType,
                        name: str,
                        const_labels: LabelsType) -> List[str]:
        '''
        :param summary: a 2-tuple containing labels and a dict representing
          the summary value. The dict contains keys for each quantile as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        summary_labels, summary_value_dict = summary
        # typing check, no runtime behaviour.
        summary_value_dict = cast(SummaryDictType, summary_value_dict)
        results = []  # type: List[str]

        for k, v in summary_value_dict.items():
            # Start from a fresh dict for the labels (new or with preset data)
            labels = {}  # type: Dict[str, str]
            if summary_labels:
                labels = summary_labels.copy()

            # Quantiles need labels and not special name (like sum and count)
            if type(k) is not float:
                name_str = "{0}_{1}".format(name, k)
            else:
                labels['quantile'] = str(k)
                name_str = name
            results.append(
                self._format_line(name_str, labels, v, const_labels))

        return results

    def _format_histogram(self,
                          histogram: MetricTupleType,
                          name: str,
                          const_labels: LabelsType) -> List[str]:
        '''
        :param histogram: a 2-tuple containing labels and a dict representing
          the histogram value. The dict contains keys for each bucket as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        histogram_labels, histogram_value_dict = histogram
        # typing check, no runtime behaviour.
        histogram_value_dict = cast(HistogramDictType, histogram_value_dict)
        results = []  # type: List[str]

        for k, v in histogram_value_dict.items():
            # Stat from a fresh dict for the labels (new or with preset data)
            labels = {}  # type: Dict[str, str]
            if histogram_labels:
                labels = histogram_labels.copy()

            # Buckets need labels and not special name (like sum and count)
            if type(k) is not float:
                name_str = "{0}_{1}".format(name, k)
            else:
                upper_bound = k
                if upper_bound == POS_INF:
                    upper_bound = '+Inf'
                elif upper_bound == NEG_INF:
                    upper_bound = '-Inf'
                # Add the le ("less or equal") label.
                labels['le'] = str(upper_bound)
                # Use the special bucket label name
                name_str = name + '_bucket'
            results.append(
                self._format_line(name_str, labels, v, const_labels))

        return results

    def marshall_lines(self, collector: CollectorsType) -> List[str]:
        '''
        Marshalls a collector into a sequence of strings representing
        the metrics in the collector.

        :return: a list of strings.
        '''
        exec_method = None  # type: FormatterFuncType
        if isinstance(collector, Counter):
            exec_method = self._format_counter
        elif isinstance(collector, Gauge):
            exec_method = self._format_gauge
        elif isinstance(collector, Summary):
            exec_method = self._format_summary
        elif isinstance(collector, Histogram):
            exec_method = self._format_histogram
        else:
            raise TypeError("Not a valid object format")

        # create headers
        help_header = HELP_FMT.format(
            name=collector.name, doc=collector.doc)

        type_header = TYPE_FMT.format(
            name=collector.name, kind=collector.kind.name)

        # Prepare start headers
        lines = [help_header, type_header]

        for i in collector.get_all():
            i = cast(MetricTupleType, i)  # typing check, no runtime behaviour.
            r = exec_method(i, collector.name, collector.const_labels)
            lines.extend(r)

        return lines

    def marshall_collector(self, collector: CollectorsType) -> str:
        '''
        Marshalls a collector into a string containing one or more lines
        '''
        # need sort?
        result = sorted(self.marshall_lines(collector))
        return LINE_SEPARATOR_FMT.join(result)

    def marshall(self, registry: CollectorRegistry) -> bytes:
        ''' Marshalls a registry (containing collectors) into a bytes
        object '''

        blocks = []
        for i in registry.get_all():
            blocks.append(self.marshall_collector(i))

        # Sort? used in tests
        blocks = sorted(blocks)

        # Needs EOF
        blocks.append("")

        return LINE_SEPARATOR_FMT.join(blocks).encode('utf-8')




import asyncio

import aiohttp

from aioprometheus import Counter, Pusher, Registry
from aioprometheus.test_utils import AsyncioTestCase


TEST_PORT = 61423
TEST_HOST = "127.0.0.1"
TEST_URL = "http://{host}:{port}".format(host=TEST_HOST, port=TEST_PORT)


class TestPusherServer(object):
    ''' This fixture class acts as the Push Gateway.

    It handles requests and stores various request attributes in the
    test_results attribute which is later checked in tests.
    '''

    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.test_results = None

    async def handler(self, request):
        data = await request.read()
        self.test_results = {
            'path': request.path,
            'headers': request.raw_headers,
            'method': request.method,
            'body': data}
        resp = aiohttp.web.Response(status=200)
        return resp

    async def start(self, addr='', port=TEST_PORT):
        self._svc = aiohttp.web.Application()
        self._svc.router.add_route(
            '*',
            "/metrics/job/{job}",
            self.handler)
        self._handler = self._svc.make_handler()
        self._svr = await self.loop.create_server(
            self._handler, addr, port)

    async def stop(self):
        self._svr.close()
        await self._svr.wait_closed()
        await self._svc.shutdown()
        await self._handler.finish_connections(1.0)
        await self._svc.cleanup()
        self._svr = None
        self._svc = None
        self._handler = None


def expected_job_path(job):
    return Pusher.PATH.format(job)


class TestPusher(AsyncioTestCase):

    async def setUp(self):
        self.server = TestPusherServer(loop=self.loop)
        await self.server.start()

    async def tearDown(self):
        await self.server.stop()

    async def test_push_job_ping(self):
        job_name = "my-job"
        p = Pusher(job_name, TEST_URL, loop=self.loop)
        registry = Registry()
        c = Counter("total_requests", "Total requests.", {})
        registry.register(c)

        c.inc({'url': "/p/user", })

        # Push to the pushgateway
        resp = await p.replace(registry)
        self.assertEqual(resp.status, 200)

        self.assertEqual(
            expected_job_path(job_name),
            self.server.test_results['path'])

    async def test_push_add(self):
        job_name = "my-job"
        p = Pusher(job_name, TEST_URL)
        registry = Registry()
        counter = Counter("counter_test", "A counter.", {'type': "counter"})
        registry.register(counter)

        counter_data = (
            ({'c_sample': '1', 'c_subsample': 'b'}, 400),
        )

        [counter.set(c[0], c[1]) for c in counter_data]
        valid_result = (b'[\n\x0ccounter_test\x12\nA counter.\x18\x00"=\n\r'
                        b'\n\x08c_sample\x12\x011\n\x10\n\x0bc_subsample\x12'
                        b'\x01b\n\x0f\n\x04type\x12\x07counter\x1a\t\t\x00'
                        b'\x00\x00\x00\x00\x00y@')

        # Push to the pushgateway
        resp = await p.add(registry)
        self.assertEqual(resp.status, 200)

        self.assertEqual(
            expected_job_path(job_name),
            self.server.test_results['path'])
        self.assertEqual("POST", self.server.test_results['method'])
        self.assertEqual(valid_result, self.server.test_results['body'])

    async def test_push_replace(self):
        job_name = "my-job"
        p = Pusher(job_name, TEST_URL)
        registry = Registry()
        counter = Counter("counter_test", "A counter.", {'type': "counter"})
        registry.register(counter)

        counter_data = (
            ({'c_sample': '1', 'c_subsample': 'b'}, 400),
        )

        [counter.set(c[0], c[1]) for c in counter_data]
        valid_result = (b'[\n\x0ccounter_test\x12\nA counter.\x18\x00"=\n\r'
                        b'\n\x08c_sample\x12\x011\n\x10\n\x0bc_subsample\x12'
                        b'\x01b\n\x0f\n\x04type\x12\x07counter\x1a\t\t\x00'
                        b'\x00\x00\x00\x00\x00y@')

        # Push to the pushgateway
        resp = await p.replace(registry)
        self.assertEqual(resp.status, 200)

        self.assertEqual(
            expected_job_path(job_name),
            self.server.test_results['path'])
        self.assertEqual("PUT", self.server.test_results['method'])
        self.assertEqual(valid_result, self.server.test_results['body'])

    async def test_push_delete(self):
        job_name = "my-job"
        p = Pusher(job_name, TEST_URL)
        registry = Registry()
        counter = Counter("counter_test", "A counter.", {'type': "counter"})
        registry.register(counter)

        counter_data = (
            ({'c_sample': '1', 'c_subsample': 'b'}, 400),
        )

        [counter.set(c[0], c[1]) for c in counter_data]
        valid_result = (b'[\n\x0ccounter_test\x12\nA counter.\x18\x00"=\n'
                        b'\r\n\x08c_sample\x12\x011\n\x10\n\x0bc_subsample'
                        b'\x12\x01b\n\x0f\n\x04type\x12\x07counter\x1a\t\t'
                        b'\x00\x00\x00\x00\x00\x00y@')

        # Push to the pushgateway
        resp = await p.delete(registry)
        self.assertEqual(resp.status, 200)

        self.assertEqual(
            expected_job_path(job_name),
            self.server.test_results['path'])
        self.assertEqual("DELETE", self.server.test_results['method'])
        self.assertEqual(valid_result, self.server.test_results['body'])




import abc
import collections
import datetime

from typing import Dict

# typing aliases
LabelsType = Dict[str, str]


class IFormatter(abc.ABC):

    @abc.abstractmethod
    def get_headers(self):
        ''' Returns a dict of headers for this response format '''

    @abc.abstractmethod
    def _format_counter(self, counter, name, const_labels):
        '''
        Returns a representation of a counter value in the implemented
        format.
        :param counter: a 2-tuple containing labels and the counter value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''

    @abc.abstractmethod
    def _format_gauge(self, gauge, name, const_labels):
        '''
        Returns a representation of a gauge value in the implemented
        format.
        :param gauge: a 2-tuple containing labels and the gauge value.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''

    @abc.abstractmethod
    def _format_summary(self, summary, name, const_labels):
        '''
        Returns a representation of a summary value in the implemented
        format.

        :param summary: a 2-tuple containing labels and a dict representing
          the summary value. The dict contains keys for each quantile as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''

    @abc.abstractmethod
    def _format_histogram(self, histogram, name, const_labels):
        '''
        Returns a representation of a histogram value in the implemented
        format.

        :param histogram: a 2-tuple containing labels and a dict representing
          the histogram value. The dict contains keys for each bucket as
          well as the sum and count fields.
        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''

    @abc.abstractmethod
    def marshall(self, registry) -> bytes:
        ''' Marshalls a registry (containing many collectors) into a
        specific format.

        :returns: bytes
        '''

    def _unify_labels(self,
                      labels: LabelsType,
                      const_labels: LabelsType,
                      ordered: bool = False) -> LabelsType:
        '''
        Return a dict of all labels for a metric. This combines the explicit
        labels and any constant labels. If ordered is True then the labels
        are sorted by key.

        :param labels: a dict of labels for a metric.
        :param const_labels: a dict of constant labels to be associated with
          the metric.
        '''
        if const_labels:
            result = const_labels.copy()
            if labels:
                # Add labels to const labels
                for k, v in labels.items():
                    result[k] = v
        else:
            result = labels

        if ordered and result:
            result = collections.OrderedDict(
                sorted(result.items(), key=lambda t: t[0]))

        return result

    def _get_timestamp(self) -> int:
        '''
        Return a timestamp that can be used by a metric formatter.
        '''
        return int(
            datetime.datetime.now(
                tz=datetime.timezone.utc).timestamp() * 1000)




import collections
import enum
import json
import re

import quantile

from . import histogram
from .metricdict import MetricDict
from typing import cast, Any, Dict, List, Sequence, Tuple, Union

# Used to return the value ordered (not necessary but for consistency useful)
# type annotations are not correct for this package yet.
decoder = json.JSONDecoder(object_pairs_hook=collections.OrderedDict)  # type: ignore


METRIC_NAME_RE = re.compile(r'^[a-zA-Z_:][a-zA-Z0-9_:]*$')
RESTRICTED_LABELS_NAMES = ('job',)
RESTRICTED_LABELS_PREFIXES = ('__',)

POS_INF = float("inf")
NEG_INF = float("-inf")

# typing aliases
LabelsType = Dict[str, str]
NumericValueType = Union[int, float, histogram.Histogram, quantile.Estimator]
ValueType = Union[str, NumericValueType]
CollectorsType = Union['Counter', 'Gauge', 'Histogram', 'Summary']


class MetricsTypes(enum.Enum):
    counter = 0
    gauge = 1
    summary = 2
    untyped = 3
    histogram = 4


class Collector(object):
    ''' Base class for all collectors.

    **Metric names and labels**

    Every time series is uniquely identified by its metric name and a set
    of key-value pairs, also known as labels.

    Labels enable Prometheus's dimensional data model: any given combination
    of labels for the same metric name identifies a particular dimensional
    instantiation of that metric (for example: all HTTP requests that used
    the method POST to the /api/tracks handler). The query language allows
    filtering and aggregation based on these dimensions. Changing any label
    value, including adding or removing a label, will create a new time
    series.

    Label names may contain ASCII letters, numbers, as well as underscores.
    They must match the regex ``[a-zA-Z_][a-zA-Z0-9_]*``. Label names
    beginning with ``__`` are reserved for internal use.

    **Samples**

    Samples form the actual time series data. Each sample consists of:

      - a float64 value
      - a millisecond-precision timestamp

    **Notation**

    Given a metric name and a set of labels, time series are frequently
    identified using this notation:

    .. code-block:: console

        <metric name>{<label name>=<label value>, ...}

    For example, a time series with the metric name
    ``api_http_requests_total`` and the labels ``method="POST"`` and
    ``handler="/messages"`` could be written like this:

    .. code-block:: console

        api_http_requests_total{method="POST", handler="/messages"}

    '''

    kind = MetricsTypes.untyped

    def __init__(self,
                 name: str,
                 doc: str,
                 const_labels: LabelsType = None) -> None:
        if not METRIC_NAME_RE.match(name):
            raise ValueError('Invalid metric name: {}'.format(name))
        self.name = name
        self.doc = doc
        self.const_labels = const_labels

        if const_labels:
            self._label_names_correct(const_labels)
            self.const_labels = const_labels

        self.values = MetricDict()

    def set_value(self, labels: LabelsType, value: NumericValueType) -> None:
        '''  Sets a value in the container '''
        if labels:
            self._label_names_correct(labels)
        self.values[labels] = value

    def get_value(self, labels: LabelsType) -> NumericValueType:
        '''  Gets a value in the container.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return self.values[labels]

    def get(self, labels: LabelsType) -> NumericValueType:
        ''' Gets a value in the container.

        Handy alias for `get_value`.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return self.get_value(labels)

    def _label_names_correct(self, labels: LabelsType) -> bool:
        ''' Check validity of label names.

        :raises: ValueError if labels are invalid
        '''
        for k, v in labels.items():
            # Check reserved labels
            if k in RESTRICTED_LABELS_NAMES:
                raise ValueError("Invalid label name: {}".format(k))

            if self.kind == MetricsTypes.histogram:
                if k in ('le', ):
                    raise ValueError("Invalid label name: {}".format(k))

            # Check prefixes
            if any(k.startswith(i) for i in RESTRICTED_LABELS_PREFIXES):
                raise ValueError("Invalid label prefix: {}".format(k))

        return True

    def get_all(self) -> List[Tuple[LabelsType, NumericValueType]]:
        '''
        Returns a list populated with 2-tuples. The first element is
        a dict of labels and the second element is the value of the metric
        itself.
        '''
        items = self.values.items()

        result = []
        for k, v in items:
            # Check if is a single value dict (custom empty key)
            if not k or k == MetricDict.EMPTY_KEY:
                key = None
            else:
                key = decoder.decode(k)
            result.append((key, self.get(k)))

        return result

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, self.__class__) and
            self.name == other.name and
            self.doc == other.doc and
            type(self) == type(other) and
            self.values == other.values)


class Counter(Collector):
    '''
    A counter is a cumulative metric that represents a single numerical value
    that only ever goes up. A counter is typically used to count requests
    served, tasks completed, errors occurred, etc. Counters should not be used
    to expose current counts of items whose number can also go down, e.g. the
    number of currently running coroutines. Use gauges for this use case.

    Examples:
    - Number of requests processed
    - Number of items that were inserted into a queue
    - Total amount of data that a system has processed
    '''

    kind = MetricsTypes.counter

    def get(self, labels: LabelsType) -> NumericValueType:
        ''' Get gets the Counter value matching an arbitrary group of labels.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return self.get_value(labels)

    def set(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Set is used to set the Counter to an arbitrary value. '''
        self.set_value(labels, value)

    def inc(self, labels: LabelsType) -> None:
        ''' Inc increments the counter by 1.'''
        self.add(labels, 1)

    def add(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Add will add the given value to the counter.

        :raises: ValueError if the value is negative. Counters can only
          increase.
        '''
        value = cast(Union[float, int], value)  # typing check, no runtime behaviour.
        if value < 0:
            raise ValueError("Counters can't decrease")

        try:
            current = self.get_value(labels)
        except KeyError:
            current = 0

        current = cast(Union[float, int], current)  # typing check, no runtime behaviour.
        self.set_value(labels, current + value)


class Gauge(Collector):
    '''
    A Gauge is a metric that represents a single numerical value that can
    arbitrarily go up and down.

    Examples of Gauges include:
    - Inprogress requests
    - Number of items in a queue
    - Free memory
    - Total memory
    - Temperature

     Gauges can go both up and down.
    '''

    kind = MetricsTypes.gauge

    def set(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Set sets the Gauge to an arbitrary value.'''
        self.set_value(labels, value)

    def get(self, labels: LabelsType) -> NumericValueType:
        ''' Get gets the Gauge value matching an arbitrary group of labels.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return self.get_value(labels)

    def inc(self, labels: LabelsType) -> None:
        ''' Inc increments the Gauge by 1.'''
        self.add(labels, 1)

    def dec(self, labels: LabelsType) -> None:
        ''' Dec decrements the Gauge by 1.'''
        self.add(labels, -1)

    def add(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Add adds the given value to the Gauge. (The value can be
            negative, resulting in a decrease of the Gauge.)
        '''
        value = cast(Union[float, int], value)  # typing check, no runtime behaviour.

        try:
            current = self.get_value(labels)
        except KeyError:
            current = 0
        current = cast(Union[float, int], current)  # typing check, no runtime behaviour.

        self.set_value(labels, current + value)

    def sub(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Sub subtracts the given value from the Gauge. (The value can be
            negative, resulting in an increase of the Gauge.)
        '''
        value = cast(Union[float, int], value)  # typing check, no runtime behaviour.
        self.add(labels, -value)


class Summary(Collector):
    '''
    A Summary captures individual observations from an event or sample stream
    and summarizes them in a manner similar to traditional summary statistics:

    1. sum of observations,
    2. observation count,
    3. rank estimations.

    Example use cases for Summaries:
    - Response latency
    - Request size
    '''

    kind = MetricsTypes.summary

    REPR_STR = "summary"
    DEFAULT_INVARIANTS = [(0.50, 0.05), (0.90, 0.01), (0.99, 0.001)]
    SUM_KEY = "sum"
    COUNT_KEY = "count"

    def __init__(self,
                 name: str,
                 doc: str,
                 const_labels: LabelsType = None,
                 invariants: Sequence[Tuple[float, float]] = DEFAULT_INVARIANTS) -> None:
        super().__init__(name, doc, const_labels=const_labels)
        self.invariants = invariants

    def add(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Add adds a single observation to the summary '''

        value = cast(Union[float, int], value)  # typing check, no runtime behaviour.
        if type(value) not in (float, int):
            raise TypeError("Summary only works with digits (int, float)")

        try:
            e = self.get_value(labels)
        except KeyError:
            # Initialize quantile estimator
            e = quantile.Estimator(*self.invariants)
            self.set_value(labels, e)

        e.observe(float(value))  # type: ignore

    # https://prometheus.io/docs/instrumenting/writing_clientlibs/#summary
    # A summary MUST have the ``observe`` methods
    observe = add

    def get(self,
            labels: LabelsType) -> Dict[Union[float, str], NumericValueType]:
        '''
        Get gets a dict of values, containing the sum, count and percentiles,
        matching an arbitrary group of labels.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return_data = {}  # type: Dict[Union[float, str], NumericValueType]

        e = self.get_value(labels)
        e = cast(Any, e)  # typing check, no runtime behaviour.

        # Set invariants data (default to 0.50, 0.90 and 0.99)
        for i in e._invariants:  # type: ignore
            q = i._quantile
            return_data[q] = e.query(q)  # type: ignore

        # Set sum and count
        return_data[self.SUM_KEY] = e._sum  # type: ignore
        return_data[self.COUNT_KEY] = e._observations  # type: ignore

        return return_data


class Histogram(Collector):
    '''
    A Histogram tracks the size and number of events in buckets.

    You can use Histograms for aggregatable calculation of quantiles.

    Example use cases:
    - Response latency
    - Request size
    '''

    kind = MetricsTypes.histogram

    REPR_STR = "histogram"
    DEFAULT_BUCKETS = (0.005, 0.01, 0.025, 0.05, 0.1, 0.25,
                       0.5, 1.0, 2.5, 5.0, 10.0, POS_INF)
    SUM_KEY = "sum"
    COUNT_KEY = "count"

    def __init__(self,
                 name: str,
                 doc: str,
                 const_labels: LabelsType = None,
                 buckets: Sequence[float] = DEFAULT_BUCKETS) -> None:
        super().__init__(name, doc, const_labels=const_labels)
        self.upper_bounds = buckets

    def add(self, labels: LabelsType, value: NumericValueType) -> None:
        ''' Add adds a single observation to the histogram '''

        value = cast(Union[float, int], value)  # typing check, no runtime behaviour.
        if type(value) not in (float, int):
            raise TypeError("Histogram only works with digits (int, float)")

        try:
            h = self.get_value(labels)
            h = cast(histogram.Histogram, h)  # typing check, no runtime behaviour.
        except KeyError:
            # Initialize histogram aggregator
            h = histogram.Histogram(*self.upper_bounds)
            self.set_value(labels, h)

        h.observe(float(value))

    # https://prometheus.io/docs/instrumenting/writing_clientlibs/#histogram
    # A histogram MUST have the ``observe`` methods
    observe = add

    def get(self,
            labels: LabelsType) -> Dict[Union[float, str], NumericValueType]:
        '''
        Get gets a dict of values, containing the sum, count and buckets,
        matching an arbitrary group of labels.

        :raises: KeyError if an item with matching labels is not present.
        '''
        return_data = {}  # type: Dict[Union[float, str], NumericValueType]

        h = self.get_value(labels)
        h = cast(histogram.Histogram, h)  # typing check, no runtime behaviour.

        for upper_bound, cumulative_count in h.buckets.items():
            return_data[upper_bound] = cumulative_count  # keys are floats

        # Set sum and count
        return_data[self.SUM_KEY] = h.sum
        return_data[self.COUNT_KEY] = h.observations

        return return_data




from collections import OrderedDict
from typing import List, Sequence, Union

POS_INF = float("inf")
NEG_INF = float("-inf")

# typing aliases
BucketType = float


def linearBuckets(start: Union[float, int],
                  width: Union[int, float],
                  count: int) -> List[BucketType]:
    '''
    Returns 'count' buckets, each 'width' wide, where the lowest bucket has an
    upper bound of 'start'. No +Inf bucket is included in the returned list.

    Raises exception if 'count' is zero or negative.
    '''
    if count < 1:
        raise Exception('Invalid count, must be a positive number')
    return [start + i * width for i in range(count)]


def exponentialBuckets(start: Union[float, int],
                       factor: Union[float, int],
                       count: int) -> List[BucketType]:
    '''
    Returns 'count' buckets, where the lowest bucket has an upper bound of
    'start' and each following bucket's upper bound is 'factor' times the
    previous bucket's upper bound. No +Inf bucket is included in the returned
    list.

    Raises exception if 'count' is 0 or negative, if 'start' is 0 or negative,
    or if 'factor' is less than or equal 1.
    '''

    if count < 1:
        raise Exception('Invalid count, must be a positive number')
    if start <= 0:
        raise Exception('Invalid start, must be positive')
    if factor < 1:
        raise Exception('Invalid factor, must be greater than one')
    return [start * (i * factor) for i in range(count)]


class Histogram(object):
    '''
    A Histogram counts individual observations from an event into configurable
    buckets. This histogram implementation also provides a sum and count of
    observations.
    '''

    def __init__(self, *buckets: BucketType) -> None:
        _buckets = [float(b) for b in buckets]

        if _buckets != sorted(buckets):
            raise ValueError('Buckets not in sorted order')

        if _buckets and _buckets[-1] != POS_INF:
            _buckets.append(POS_INF)

        if len(_buckets) < 2:
            raise ValueError('Must have at least two buckets')

        self.buckets = OrderedDict(
            [(b, 0) for b in _buckets])  # type: Dict[float, int]
        self.observations = 0  # type: int
        self.sum = 0.0  # type: float

    def observe(self, value: Union[float, int]) -> None:
        ''' Observe the given amount.

        Increment the count of observations, add value to the sum and
        increment the appropriate bucket counter.
        '''
        # The last bucket is +Inf, so we will always increment at least
        # one bucket
        for upper_bound in self.buckets:
            if value <= upper_bound:
                self.buckets[upper_bound] += 1
                break
        self.sum += value
        self.observations += 1




import unittest

from aioprometheus import (
    Collector,
    Counter,
    Gauge,
    Histogram,
    Registry,
    Summary)


class TestCollectorDict(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

        self.c = Collector(**self.data)

    def test_initialization(self):
        self.assertEqual(self.data['name'], self.c.name)
        self.assertEqual(self.data['doc'], self.c.doc)
        self.assertEqual(self.data['const_labels'], self.c.const_labels)

    def test_set_value(self):
        data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
        )

        for m in data:
            self.c.set_value(m[0], m[1])

        self.assertEqual(len(data), len(self.c.values))

    def test_same_value(self):
        data = (
            ({'country': "sp", "device": "desktop", "ts": "GMT+1"}, 520),
            ({"ts": "GMT+1", 'country': "sp", "device": "desktop"}, 521),
            ({'country': "sp", "ts": "GMT+1", "device": "desktop"}, 522),
            ({"device": "desktop", "ts": "GMT+1", 'country': "sp"}, 523),
        )

        for m in data:
            self.c.set_value(m[0], m[1])

        self.assertEqual(1, len(self.c.values))
        self.assertEqual(523, self.c.values[data[0][0]])

    def test_get_value(self):
        data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
        )

        for m in data:
            self.c.set_value(m[0], m[1])

        for m in data:
            self.assertEqual(m[1], self.c.get_value(m[0]))

    def test_not_const_labels(self):
        del self.data['const_labels']
        self.c = Collector(**self.data)

    def test_not_name(self):
        with self.assertRaises(TypeError) as context:
            del self.data['name']
            self.c = Collector(**self.data)

        self.assertEqual(
            "__init__() missing 1 required positional argument: 'name'",
            str(context.exception))

    def test_not_help_text(self):
        with self.assertRaises(TypeError) as context:
            del self.data['doc']
            self.c = Collector(**self.data)

        self.assertEqual(
            "__init__() missing 1 required positional argument: 'doc'",
            str(context.exception))

    def test_without_labels(self):
        data = (
            ({}, 520),
            (None, 654),
            ("", 1001),
        )

        for i in data:
            self.c.set_value(i[0], i[1])

        self.assertEqual(1, len(self.c.values))
        self.assertEqual((data)[len(data) - 1][1], self.c.values[data[0][0]])

    def test_wrong_labels(self):

        # Normal set
        with self.assertRaises(ValueError) as context:
            self.c.set_value({'job': 1, 'ok': 2}, 1)

        self.assertEqual('Invalid label name: job', str(context.exception))

        with self.assertRaises(ValueError) as context:
            self.c.set_value({'__not_ok': 1, 'ok': 2}, 1)

        self.assertEqual('Invalid label prefix: __not_ok', str(context.exception))

        # Constructor set
        with self.assertRaises(ValueError) as context:
            Collector("x", "y", {'job': 1, 'ok': 2})

        self.assertEqual('Invalid label name: job', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Collector("x", "y", {'__not_ok': 1, 'ok': 2})

        self.assertEqual('Invalid label prefix: __not_ok', str(context.exception))

    def test_get_all(self):
        data = (
            ({'country': "sp", "device": "desktop"}, 520),
            ({'country': "us", "device": "mobile"}, 654),
            ({'country': "uk", "device": "desktop"}, 1001),
            ({'country': "de", "device": "desktop"}, 995),
            ({'country': "zh", "device": "desktop"}, 520),
            ({'country': "ch", "device": "mobile"}, 654),
            ({'country': "ca", "device": "desktop"}, 1001),
            ({'country': "jp", "device": "desktop"}, 995),
            ({'country': "au", "device": "desktop"}, 520),
            ({'country': "py", "device": "mobile"}, 654),
            ({'country': "ar", "device": "desktop"}, 1001),
            ({'country': "pt", "device": "desktop"}, 995),
        )

        for i in data:
            self.c.set_value(i[0], i[1])

        def country_fetcher(x):
            return x[0]['country']
        sorted_data = sorted(data, key=country_fetcher)
        sorted_result = sorted(self.c.get_all(), key=country_fetcher)
        self.assertEqual(sorted_data, sorted_result)


class TestCounter(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "logged_users_total",
            'doc': "Logged users in the application",
            'const_labels': {"app": "my_app"},
        }

        self.c = Counter(**self.data)

    def test_set(self):

        data = (
            {
                'labels': {'country': "sp", "device": "desktop"},
                'values': range(10)
            },
            {
                'labels': {'country': "us", "device": "mobile"},
                'values': range(10, 20)
            },
            {
                'labels': {'country': "uk", "device": "desktop"},
                'values': range(20, 30)
            }
        )

        for i in data:
            for j in i['values']:
                self.c.set(i['labels'], j)

        self.assertEqual(len(data), len(self.c.values))

    def test_get(self):
        data = (
            {
                'labels': {'country': "sp", "device": "desktop"},
                'values': range(10)
            },
            {
                'labels': {'country': "us", "device": "mobile"},
                'values': range(10, 20)
            },
            {
                'labels': {'country': "uk", "device": "desktop"},
                'values': range(20, 30)
            }
        )

        for i in data:
            for j in i['values']:
                self.c.set(i['labels'], j)
                self.assertEqual(j, self.c.get(i['labels']))

        # Last check
        for i in data:
            self.assertEqual(max(i['values']), self.c.get(i['labels']))

    def test_set_get_without_labels(self):
        data = {
            'labels': {},
            'values': range(100)
        }

        for i in data['values']:
            self.c.set(data['labels'], i)

        self.assertEqual(1, len(self.c.values))

        self.assertEqual(max(data['values']), self.c.get(data['labels']))

    def test_inc(self):
        iterations = 100
        labels = {'country': "sp", "device": "desktop"}

        for i in range(iterations):
            self.c.inc(labels)

        self.assertEqual(iterations, self.c.get(labels))

    def test_add(self):
        labels = {'country': "sp", "device": "desktop"}
        iterations = 100

        for i in range(iterations):
            self.c.add(labels, i)

        self.assertEqual(sum(range(iterations)), self.c.get(labels))

    def test_negative_add(self):
        labels = {'country': "sp", "device": "desktop"}

        with self.assertRaises(ValueError) as context:
            self.c.add(labels, -1)
        self.assertEqual('Counters can\'t decrease', str(context.exception))


class TestGauge(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "hdd_disk_used",
            'doc': "Disk space used",
            'const_labels': {"server": "1.db.production.my-app"},
        }

        self.g = Gauge(**self.data)

    def test_set(self):
        data = (
            {
                'labels': {'max': "500G", 'dev': "sda"},
                'values': range(0, 500, 50)
            },
            {
                'labels': {'max': "1T", 'dev': "sdb"},
                'values': range(0, 1000, 100)
            },
            {
                'labels': {'max': "10T", 'dev': "sdc"},
                'values': range(0, 10000, 1000)
            }
        )

        for i in data:
            for j in i['values']:
                self.g.set(i['labels'], j)

        self.assertEqual(len(data), len(self.g.values))

    def test_get(self):
        data = (
            {
                'labels': {'max': "500G", 'dev': "sda"},
                'values': range(0, 500, 50)
            },
            {
                'labels': {'max': "1T", 'dev': "sdb"},
                'values': range(0, 1000, 100)
            },
            {
                'labels': {'max': "10T", 'dev': "sdc"},
                'values': range(0, 10000, 1000)
            }
        )

        for i in data:
            for j in i['values']:
                self.g.set(i['labels'], j)
                self.assertEqual(j, self.g.get(i['labels']))

        for i in data:
            self.assertEqual(max(i['values']), self.g.get(i['labels']))

    def test_set_get_without_labels(self):
        data = {
            'labels': {},
            'values': range(100)
        }

        for i in data['values']:
            self.g.set(data['labels'], i)

        self.assertEqual(1, len(self.g.values))

        self.assertEqual(max(data['values']), self.g.get(data['labels']))

    def test_inc(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}

        for i in range(iterations):
            self.g.inc(labels)
            self.assertEqual(i + 1, self.g.get(labels))

        self.assertEqual(iterations, self.g.get(labels))

    def test_dec(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}
        self.g.set(labels, iterations)

        for i in range(iterations):
            self.g.dec(labels)
            self.assertEqual(iterations - (i + 1), self.g.get(labels))

        self.assertEqual(0, self.g.get(labels))

    def test_add(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}

        for i in range(iterations):
            self.g.add(labels, i)

        self.assertEqual(sum(range(iterations)), self.g.get(labels))

    def test_add_negative(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}

        for i in range(iterations):
            self.g.add(labels, -i)

        self.assertEqual(sum(map(lambda x: -x, range(iterations))),
                         self.g.get(labels))

    def test_sub(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}

        for i in range(iterations):
            self.g.sub(labels, i)

        self.assertEqual(sum(map(lambda x: -x, range(iterations))),
                         self.g.get(labels))

    def test_sub_positive(self):
        iterations = 100
        labels = {'max': "10T", 'dev': "sdc"}

        for i in range(iterations):
            self.g.sub(labels, -i)

        self.assertEqual(sum(range(iterations)), self.g.get(labels))


class TestSummary(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "http_request_duration_microseconds",
            'doc': "Request duration per application",
            'const_labels': {"app": "my_app"},
        }

        self.s = Summary(**self.data)

    def test_add(self):
        data = (
            {
                'labels': {'handler': '/static'},
                'values': range(0, 500, 50)
            },
            {
                'labels': {'handler': '/p'},
                'values': range(0, 1000, 100)
            },
            {
                'labels': {'handler': '/p/login'},
                'values': range(0, 10000, 1000)
            }
        )

        for i in data:
            for j in i['values']:
                self.s.add(i['labels'], j)

        for i in data:
            self.assertEqual(len(i['values']),
                             self.s.values[i['labels']]._observations)

    def test_get(self):
        labels = {'handler': '/static'}
        values = [3, 5.2, 13, 4]

        for i in values:
                self.s.add(labels, i)

        data = self.s.get(labels)
        correct_data = {
            'sum': 25.2,
            'count': 4,
            0.50: 4.0,
            0.90: 5.2,
            0.99: 5.2,
        }

        self.assertEqual(correct_data, data)

    def test_add_get_without_labels(self):
        labels = None
        values = [3, 5.2, 13, 4]

        for i in values:
            self.s.add(labels, i)

        self.assertEqual(1, len(self.s.values))

        correct_data = {
            'sum': 25.2,
            'count': 4,
            0.50: 4.0,
            0.90: 5.2,
            0.99: 5.2,
        }
        self.assertEqual(correct_data, self.s.get(labels))

    def test_add_wrong_types(self):
        labels = None
        values = ["3", (1, 2), {'1': 2}, True]

        for i in values:
            with self.assertRaises(TypeError) as context:
                self.s.add(labels, i)
        self.assertEqual("Summary only works with digits (int, float)",
                         str(context.exception))


class TestHistogram(unittest.TestCase):

    def setUp(self):
        self.data = {
            'name': "http_request_duration_microseconds",
            'doc': "Request duration per application",
            'const_labels': {"app": "my_app"},
        }

        self.h = Histogram(**self.data)

    def test_wrong_labels(self):

        with self.assertRaises(ValueError) as context:
            self.h.set_value({'le': 2}, 1)

        self.assertEqual('Invalid label name: le', str(context.exception))




import os
import re

from pip.req import parse_requirements
from pip.download import PipSession
from setuptools import setup, find_packages


install_reqs = parse_requirements("requirements.txt", session=PipSession())
requires = [str(ir.req) for ir in install_reqs]


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*['\"](\d\d\.\d\d\.\d+)['\"]")
    init_file = os.path.join(
        os.path.dirname(__file__), 'aioprometheus', '__init__.py')
    with open(init_file) as f:
        for line in f:
            match = regexp.match(line)
            if match:
                return match.group(1)
        else:
            raise RuntimeError(
                'Cannot find __version__ in aioprometheus/__init__.py')

version = read_version()


if __name__ == "__main__":

    setup(
        name="aioprometheus",
        version=version,
        author="Chris Laws",
        author_email="clawsicus@gmail.com",
        description="A Prometheus Python client library for asyncio-based applications",
        long_description="",
        license="MIT",
        keywords=["prometheus", "monitoring", "metrics"],
        url="https://github.com/claws/aioprometheus",
        packages=find_packages(),
        install_requires=requires,
        pyrobuf_modules="proto",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.5",
            "Topic :: System :: Monitoring"])
